## @namespace officegenerator.tests.test_demo
## @brief Test for officegenerator_demo script

import unittest
from officegenerator.demo import main


## Class to text Range operations. Class must begin with Test and modules with test_ too

class TestArgs(unittest.TestCase):
    def test_Coord_methods(self):
        main(["--create",])
        main(["--remove",])


if __name__ == '__main__':
    unittest.main()


## @brief Module with objects managers as list or as dictionary.
## THIS IS FILE IS FROM https://github.com/turulomio/reusingcode IF YOU NEED TO UPDATE IT PLEASE MAKE A PULL REQUEST IN THAT PROJECT
## DO NOT UPDATE IT IN YOUR CODE IT WILL BE REPLACED USING FUNCTION IN README
##
## You have to use list objects if you are going to make selections and secuential access.

from datetime import datetime, timedelta, date
from logging import critical, debug
from .datetime_functions import dtaware_day_end_from_date, dtaware_day_start_from_date, dtnaive_day_end_from_date, dtnaive_day_start_from_date
from .call_by_name import call_by_name

## Defines who self.selected is managed
## If can take the following values
## - Object self.selected is a object
## - List self.selected is a list of objects
## - Manager. Selef selected is an object
class ManagerSelectionMode:
    Object=0
    List=1
    Manager=2

class ObjectManager(object):
    def __init__(self):
        self.arr=[]

    def __repr__(self):
        return "{} with {} objects".format(self.__class__.__name__, self.length())

    ## Method to iterate self.arr iterating object
    def __iter__(self):
        return iter(self.arr)

    ## Store constructor params to allow create new instances of this managers
    def setConstructorParameters(self, *params):
        self.initparams=params

    def append(self,  obj):
        self.arr.append(obj)

    ## Appends and object if it isn't in array. Array will act as a set()
    def append_distinct(self, obj):
        if obj not in self.arr:
            self.append(obj)

    def remove(self, obj):
        self.arr.remove(obj)

    ## Needs the setConstructorParameters before
    def emptyManager(self):
        return self.__class__(*self.initparams)

    def length(self):
        return len(self.arr)

    def clean(self):
        """Deletes all items"""
        self.arr=[]

    def clone(self):
        """Returns other Set object, with items referenced, ojo con las formas de las instancias
        initparams son los parametros de iniciación de la clase"""
        result=self.emptyManager()
        for a in self.arr:
            result.append(a)
        return result
   
    def first(self):
        return self.object(0)

    def index(self,o):
        return self.arr.index(o)

    def last(self):
        return self.object(self.length()-1)

    def print(self):
        print ("Objects in {}".format(self.__class__))
        for q in self.arr:
            print(" * {}".format(q))
            
    ## Return the object at the index position
    def object(self, index):
        try:
            return self.arr[index]
        except:
            critical("I couldn't retrive object from {} position".format(index))

    ## Order data columns. None values are set at the beginning
    def order_with_none(self, string_or_tuple, reverse=False, none_at_top=True):
        nonull=[]
        null=[]
        for o in self.arr:
            com=call_by_name(o, string_or_tuple)
            if com is None:
                null.append(o)
            else:
                nonull.append(o)
        nonull=sorted(nonull, key=lambda c: call_by_name(c,string_or_tuple), reverse=reverse)
        if none_at_top==True:#Set None at top of the list
            self.arr=null+nonull
        else:
            self.arr=nonull+null

    ## @param string_or_tuple String or tuple used with a call_by_name method
    ## @return List returned with a call_by_name string or tuplen array with all object ids
    def list_of(self, string_or_tuple):
        r=[]
        for o in self:
            r.append(call_by_name(o, string_or_tuple))
        return r

    ## Returns a new manager with the objects that have found a list of strings in several commands, passed as 
    ## This function doen't make exact matches. Only if strings in s_list are contained
    ## @param string_or_tuple_list List of _string_or_tuple_to_command parameters
    ## @param s_list List of string to search
    ## @param upper boolean
    def find_strings_contained_in_string_or_tuple_results(self, string_or_tuple_list, s_list, upper=False):
        r=self.emptyManager()
        for o in self.arr:
            string_=""
            for string_or_tuple in string_or_tuple_list:
                string_=string_+str(call_by_name(o, string_or_tuple))
                
            for s in s_list:
                if upper==True:
                    if s.upper() in string_.upper():
                        r.append(o)
                        break
                else:#upper False
                    if s in string_:
                        r.append(o)
                        break
        return r
        
    def find_string_exact_in_string_or_tuple_results(self, string_or_tuple, s, upper=False):        
        r=self.emptyManager()
        for o in self.arr:
            if upper==True:
                if s.upper() == call_by_name(o, string_or_tuple).upper():
                    r.append(o)
            else:#upper False
                if s == call_by_name(o, string_or_tuple):
                    r.append(o)
        return r


    ## Search by id iterating array
    def find_by_id_builtin(self, id_, logging=False):
        start=datetime.now()
        for o in self.arr:
            if id(o)==id_:
                if logging==True:
                    debug("{} took {} to find by id builtin {} with list".format(self.__class__.__name__, datetime.now()-start, id))
                return o
        return None


    ## It doesn't emit selected if selected is None a nd needtoselect is False, in the rest of the cases it emmit itemChanged
    ## @param combo
    ## @param selected it's an object
    ## @param needtoselect Adds a foo item with value==None with the text select one
    ## @param icons Boolean. If it's true uses o.qicon() method to add an icon to the item
    ## @param id_attr. Atribute to call_by_name for id. If None uses id(o). Normally will user "id"
    ## @param name_attr. Atribute to call_by_name for name. If None user str(o). Normally will use "name"
    
    def qcombobox(self, combo,  selected=None, needtoselect=False, icons=False, id_attr="id", name_attr="name"):
        combo.blockSignals(True)
        combo.clear()

        #Add items
        if needtoselect==True:
            if self.length()>0:
                combo.addItem(combo.tr("Select an option"), None)
            else:
                combo.addItem(combo.tr("No options to select"), None)
        for a in self.arr:
            print(a,id_attr,name_attr)
            id_  =call_by_name(a, id_attr) if id_attr is not None else id(a)
            name_=call_by_name(a, name_attr) if name_attr is not None else str(a)
            if icons==True:
                combo.addItem(a.qicon(), name_, id_)
            else:
                combo.addItem(name_, id_)

        #Force without signals to be in -1. There were problems when 0 is selected, becouse it didn't emit anything
        combo.setCurrentIndex(-1)

        #Set selection
        if needtoselect==True:
            combo.blockSignals(False)
            combo.setCurrentIndex(0)
        else:#need to select False
            if selected is None:
                combo.blockSignals(False)
            else:
                combo.blockSignals(False)
                if id_attr is None:
                    selected_id=id(id_attr)
                else:
                    selected_id=selected.id
                combo.setCurrentIndex(combo.findData(selected_id))


## Manager Selection class
## By default selectionmode is
class ManagerSelection(object):
    def __init__(self):
        self.__selected=None
        self.__selectionmode=ManagerSelectionMode.Object

    @property
    def selected(self):
        return self.__selected

    @selected.setter
    def selected(self, value):
        self.__selected=value
        
    def selectionMode(self):
        return self.__selectionmode

    ## @param value ManagerSelectionMode
    def setSelectionMode(self, value):
        self.__selectionmode=value
        if value==ManagerSelectionMode.Object:
            self.selected=None
        elif value==ManagerSelectionMode.List:
            self.selected=[]
        elif value==ManagerSelectionMode.Manager:
            self.selected=self.emptyManager()

    def cleanSelection(self):
        if self.selectionMode()==ManagerSelectionMode.Object:
            self.selected=None
        elif self.selectionMode()==ManagerSelectionMode.List:
            self.selected=[]
        elif self.selectionMode()==ManagerSelectionMode.Manager:
            self.selected.clean()

    ## Useful to setselection without interactivvite ui
    ## @param list Can be, list, manager or object
    def setSelected(self, list):
        self.cleanSelection()
        if self.selectionMode() in (ManagerSelectionMode.List, ManagerSelectionMode.Manager):
            for o in list:
                self.selected.append(o)
        else:#Object
            self.selected=list
    
    ## Converts selection to a manager. With this function I removed ManagerSelectionMode.Manager.
    def convertSelectionToManager(self):
        r=self.emptyManager()
        if self.selectionMode()==ManagerSelectionMode.List:
            r.arr=self.selected
        else:#Object
            r.append(self.selected)
        return r

## Objects in DictListObjectManager has and id. The Id can be a integer or a string or ...
class ObjectManager_With_Id(ObjectManager):
    def __init__(self):
        ObjectManager.__init__(self)
        self._find_dict={}
        self._use_dict_to_find=False

    ## If set to True enables the use of a dict to find by id
    def setUseDictToFind(self, value):
        self._use_dict_to_find=value

    def append(self,  obj):
        self.arr.append(obj)
        if self._use_dict_to_find==True:
            if obj.id is None:
                debug("You have added a key None to self._find_dict, perhaps you need to append it when object.id is set")
            self._find_dict[obj.id]=obj

    def remove(self, obj):
        self.arr.remove(obj)
        if self._use_dict_to_find==True:
            del self._find_dict[obj.id]


    def arr_position(self, id):
        """Returns arr position of the id, useful to select items with unittests"""
        for i, a in enumerate(self.arr):
            if a.id==id:
                return i
        return None

    ##Returns an array with all object ids
    def array_of_ids(self):
        r=[]
        for o in self.arr:
            r.append(o.id)
        return r

    ## Search by id iterating array
    def find_by_id(self, id, logging=False):
        if self._use_dict_to_find==True:
            try:
                return self._find_dict[id]
            except:
                return None
        else:
            start=datetime.now()
            for o in self.arr:
                if o.id==id:
                    if logging==True:
                        debug("{} took {} to find by id {} with list".format(self.__class__.__name__, datetime.now()-start, id))
                    return o
            return None

    def order_by_id(self, reverse=False, none_at_top=True):
        self.order_with_none("id", reverse=reverse, none_at_top=none_at_top)

    def union(self,  set,  *initparams):
        """Returns a new set, with the union comparing id
        initparams son los parametros de iniciación de la clse"""        
        resultado=self.__class__(*initparams)#Para que coja la clase del objeto que lo invoca SetProduct(self.mem), luego será self.mem
        for p in self.arr:
            resultado.append(p)
        for p in set.arr:
            if resultado.find_by_id(p.id, False)==None:
                resultado.append(p)
        return resultado


## Objects in DictListObjectManager has and id and a date attribute
class ObjectManager_With_IdDate(ObjectManager_With_Id):
    def __init__(self):
        ObjectManager_With_Id.__init__(self)

    def order_by_date(self, reverse=False, none_at_top=True):
        self.order_with_none("date", reverse=reverse, none_at_top=none_at_top)
        
## Objects in DictListObjectManager has and id and a datetime
class ObjectManager_With_IdDatetime(ObjectManager_With_Id):
    def __init__(self):
        ObjectManager_With_Id.__init__(self)

    def order_by_datetime(self, reverse=False, none_at_top=True):
        self.order_with_none("datetime", reverse=reverse, none_at_top=none_at_top)
                
    ## Function that returns the same object manager, with a pointer to the of the objects that contains from the datetime given in the parameter.
    ## For example the constuctor of InvemestOperationHomogeneous is InvesmentOperationHomogeneous(mem,investment). so to use this function you need ObjectManager_from_datetime(dt,mem,investment)
    ## @param dt datetime This function copies all object with datetime until this parameter
    def ObjectManager_from_datetime(self, dt):
        result=self.emptyManager()
        for a in self.arr:
            if a.datetime>=dt:
                result.append(a)
        return result         
        
    ## Function that returns the same object manager, with a pointer to the of the objects that contains from the datetime given in the parameter.
    ## For example the constuctor of InvemestOperationHomogeneous is InvesmentOperationHomogeneous(mem,investment). so to use this function you need ObjectManager_from_datetime(dt,mem,investment)
    ## @param datetime. This function copies all object with datetime until this parameter
    def ObjectManager_until_datetime(self, dt):
        result=self.emptyManager()
        for a in self.arr:
            if a.datetime<=dt:
                result.append(a)
        return result
        
    def ObjectManager_copy_from_datetime(self, dt):
        result=self.emptyManager()
        for a in self.arr:
            if a.datetime>=dt:
                result.append(a.copy())
        return result
        
    ## Function that returns the same object manager, but with a copy of the objects that contains until the datetime given in the parameter.
    ## For exemple the constuctor of InvemestOperationHomogeneous is InvemestOperationHomogeneous(mem,investment). so to use this function you need ObjectManager_copy_until_datetime(dt,mem,investment)
    ## @param dt datetime This function copies all object with datetime until this parameter
    def ObjectManager_copy_until_datetime(self, dt):
        result=self.emptyManager()
        for a in self.arr:
            if a.datetime<=dt:
                result.append(a.copy())
        return result

## Objects in DictListObjectManager has and id and a name
class ObjectManager_With_IdName(ObjectManager_With_Id):
    def __init__(self):
        ObjectManager_With_Id.__init__(self)
        
    ##Returns an array with all object name
    ## @param sort Boolean to sort or not the array
    ## @param nones Boolean. If True adds None and empty strings values to the list. If False it doesn't
    def array_of_names(self, sort=True, nones=False):
        r=[]
        for o in self.arr:
            if nones==False:
                if o.name==None or o.name=="":
                    continue
            r.append(o.name)
        if sort==True:
            r.sort()
        return r

    def find_by_name(self, s,  upper=False):
        return self.find_string_exact_in_string_or_tuple_results("name", s, upper)

    ## Returns another object manager of the same class with the elements that contains a string in the name
    ## @param s string to search
    ## @param casesensitive Boolean if it's a case sensitive search    
    def ObjectManager_which_name_contains(self, s, casesensitive):
        result=self.emptyManager()
        if s is None:
            critical("Search string can't be None")
            return result
        if casesensitive==True:
            for a in self.arr:
                if s in a.name:
                    result.append(a)
            return result
        else:
            for a in self.arr:
                if s.upper() in a.name.upper():
                    result.append(a)
            return result
        
    def order_by_name(self, reverse=False, none_at_top=True):
        self.order_with_none("name", reverse=reverse, none_at_top=none_at_top)

    def order_by_upper_name(self, reverse=False, none_at_top=True):
        self.order_with_none(("name.upper", []), reverse=reverse, none_at_top=none_at_top)


    ## Creates a libreoffice sheet from the ObjectManager
    ##
    ## This function needs the officegenerator package
    ## @param ods Officegenerator ODS_Write object
    ## @param sheetname String with the name of the libreoffice sheet
    ## @param titles List of strings with the titles of the columns
    ## @param order_by_name Boolean. True: orders by name. False: orders by id
    ## @returns Officegenerator OdfSheet
    def ods_sheet(self, ods, sheetname, titles=["Id", "Name"],  order_by_name=True):
        from officegenerator import Coord
        if order_by_name==True:
            self.order_by_name()
        else:
            self.order_by_id()
        s=ods.createSheet(sheetname)
        s.setColumnsWidth([80, 240])
        s.add("A1", [titles], "OrangeCenter")
        for number, o in enumerate(self.arr):
            s.add(Coord("A2").addRow(number), o.id, "WhiteRight")        
            s.add(Coord("B2").addRow(number), o.name, "WhiteLeft")
        s.freezeAndSelect("A1")
        return s

## Usefull when creating a class with two attributes self.id and self.name only
class Object_With_IdName:
    ## Constructor with the following attributes combination
    ## 1. Object_With_IdName(). Create an Object_With_IdName with all attributes to None
    ## 2. Object_With_IdName( id,  name). Create an Object_With_IdName settings all attributes.
    ## @param name String with the name of the Object_With_IdName
    ## @param id Integer that sets the id of the Object_With_IdName
    def __init__(self, id=None,  name=None):
            self.id=id
            self.name=name

    def __repr__(self):
        return "{}: {} (Id: {})".format(self.__class__.__name__, self.name, self.id)

    def upper_name(self, boolean):
        if self.name==None:
            return None
        if boolean==True:
            return self.name.upper()
        else:
            return self.name.lower()

class ObjectManager_Selectable(ObjectManager, ManagerSelection):
    def __init__(self):
        ObjectManager.__init__(self)
        ManagerSelection.__init__(self)

class ObjectManager_With_Id_Selectable(ObjectManager_With_Id, ManagerSelection):
    def __init__(self):
        ObjectManager_With_Id.__init__(self)
        ManagerSelection.__init__(self)

class ObjectManager_With_IdDate_Selectable(ObjectManager_With_IdDate, ManagerSelection):
    def __init__(self):
        ObjectManager_With_IdDate.__init__(self)
        ManagerSelection.__init__(self)

class ObjectManager_With_IdDatetime_Selectable(ObjectManager_With_IdDatetime, ManagerSelection):
    def __init__(self):
        ObjectManager_With_IdDatetime.__init__(self)
        ManagerSelection.__init__(self)

class ObjectManager_With_IdName_Selectable(ObjectManager_With_IdName, ManagerSelection):
    def __init__(self):
        ObjectManager_With_IdName.__init__(self)
        ManagerSelection.__init__(self)




## THIS IS A NEW SERIE OF MANAGERS DATETIME VALUE
class DatetimeValue:
    def __init__(self):
        self.datetime=None
        self.value=None

    def __repr__(self):
        return "DatetimeValue {} = {}".format(self.date,self.value)

class DatetimeValueManager(ObjectManager):
    def __init__(self):
        ObjectManager.__init__(self)
    def appendDV(self, datetime, value):
        o=DatetimeValue()
        o.datetime=datetime
        o.value=value
        self.append(o)    ## Returns a date value manager with the simple movil average 3 of weight
    
    ## Returns a DVManager with the simple movil average of the array
    def sma(self, period):
        r=DatetimeValueManager()
        for i in range(period, self.length()):
            sma=DatetimeValue()
            sma.value=0
            sma.datetime=self.arr[i].datetime
            for p in range(period):
                sma.value=sma.value+self.arr[i-p].value
            sma.value=sma.value/period
            r.append(sma)
        return r

    ## Return find less or equal dv of a given datetime
    ## @param dt datetime
    def find_le(self, dt):
        for o in reversed(self.arr):
            if o.datetime<=dt:
                return o
        return None

## THIS IS A NEW SERIE OF MANAGERS DATE VALUE
class DateValue:
    def __init__(self):
        self.date=None
        self.value=None

    def __repr__(self):
        return "DateValue {} = {}".format(self.date,self.value)

class DateValueManager(ObjectManager):
    def __init__(self):
        ObjectManager.__init__(self)

    def appendDV(self, date, value):
        o=DateValue()
        o.date=date
        o.value=value
        self.append(o)    ## Returns a date value manager with the simple movil average 3 of weight
        
    ## Fills days without data  with the data before or after
    def DateValueManager_filling_empty(self, with_data_before=True):
        r=DateValueManager()
        last=self.first()
        r.append(last)
        for o in self.arr[1:]:
            while o.date!=last.date+timedelta(days=1):
                missing=DateValue()
                missing.date=last.date+timedelta(days=1)
                missing.value=last.value
                r.append(missing)
                last=missing
            r.append(o)
            last=o
            
        return r
        
    ## Converts a DateValueManager to a DatetimeValueManager.
    ## @param start Boolean. If true date is converted to the start of the day. If false to the end of the day
    ## @param timezone String with pytz timexone. If None datetimes will be naive, else datetimes will be aware
    def DatetimeValueManager(self, start=True, timezone=None):
        r=DatetimeValueManager()
        for o in self.arr:
            if start==True:
                if timezone==None:
                    r.appendDV(dtnaive_day_start_from_date(o.date), o.value)
                else:
                    r.appendDV(dtaware_day_start_from_date(o.date, timezone), o.value)
            else:#end of day
                if timezone==None:
                    r.appendDV(dtnaive_day_end_from_date(o.date), o.value)
                else:
                    r.appendDV(dtaware_day_end_from_date(o.date, timezone), o.value)
        return r

    ## Returns a DVManager with the simple movil average of the array
    def sma(self, period):
        r=DateValueManager()
        for i in range(period, self.length()):
            sma=DateValue()
            sma.value=0
            sma.date=self.arr[i].date
            for p in range(period):
                sma.value=sma.value+self.arr[i-p].value
            sma.value=sma.value/period
            r.append(sma)
        return r



if __name__ == "__main__":
    print("TESTING FILLING DATEVALUEMANAGER")
    r=DateValueManager()
    r.appendDV(date(2019, 1, 10), 1)
    r.appendDV(date(2019, 1, 17), 2)
    r.print()
    filled=r.DateValueManager_filling_empty()
    filled.print()

    sizes=(1,10,100,1000,10000,100000)#,1000000,3000000)
    for size in sizes:
        l=ObjectManager_With_Id()
        for number in range(size):
            o=Object_With_IdName(number,"Name {}".format(number))
            l.append(o)
        middle=size*2//3
        start=datetime.now()
        l.find_by_id(middle)
        print("Benchmarking search_by_id in to element {} with {} objects took {}".format(middle,size,datetime.now()-start))
        
    print(l.find_by_id(500))


    print()
    print("Ordering with None")
    manager=ObjectManager_With_IdName()
    manager.setConstructorParameters()
    manager.append(Object_With_IdName(12, None))
    manager.append(Object_With_IdName(23,"AB"))
    manager.append(Object_With_IdName(34,"BC"))
    manager.append(Object_With_IdName(45,"CD"))
    manager.append(Object_With_IdName(None,"DE"))
    manager.print()
    manager.order_with_none(["upper_name", (True,)], False, False)
    manager.print()

    print()
    print("Finding objects")
    find_name=manager.find_strings_contained_in_string_or_tuple_results(
                            [
                                ("upper_name", (True, )),  #tuple
                                "id" #string
                            ],  
                            ["2", "C"] #Words to find
                        )
    find_name.print()
    find_name=manager.find_by_name("CD")
    find_name.print()

    print("Iterating an object")
    for o in manager:
        print(o)



## THIS IS FILE IS FROM https://github.com/turulomio/reusingcode IF YOU NEED TO UPDATE IT PLEASE MAKE A PULL REQUEST IN THAT PROJECT
## DO NOT UPDATE IT IN YOUR CODE IT WILL BE REPLACED USING FUNCTION IN README

from time import time
from socket import create_connection
from logging import error, debug, warning
from platform import system as platform_system
from sys import exit

## allows you to measure the execution time of the method/function by just adding the @timeit decorator on the method.
## @param method
def timeit(method):
    def timed(*args, **kw):
        ts = time()
        result = method(*args, **kw)
        te = time()

        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            debug('%r took  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result

    return timed

## Checks if there is internet
def is_there_internet(method):
    def internet(*args, **kw):
        try:
            # connect to the host -- tells us if the host is actually
            # reachable
            create_connection(("www.google.com", 80))
            debug("There is Internet connection")
        except OSError:
            error("There is not Internet connection")
        return method(*args, **kw)
    return internet

## This is a decorator which can be used to mark functions
## as deprecated. It will result in a warning being emitted
## when the function is used
def deprecated(method):
    def new_func(*args, **kwargs):
        warning("Call to deprecated function {}.".format(method.__name__))
        return method(*args, **kwargs)
    return new_func

## This function checks if currrent user is root or administrator in Windows or Linux
def need_administrator(method):
    def new_func(*args, **kwargs):
        if platform_system=='Windows':
            from ctypes.windll.shell32 import IsUserAnAdmin
            if IsUserAnAdmin()!=True:
                warning("You need to be an Administrator to execute this code.")
        else:
            from os import geteuid
            if geteuid() !=0:
                warning("You need to be root to execute this code.")
                exit(-1)
        return method(*args, **kwargs)
    return new_func


@is_there_internet
@deprecated
def testing_decorators1():
    for i in range(10):
        print(i)

@timeit
def testing_decorators2():
    for i in range(10):
        print(i)
        
if __name__ == "__main__":
    testing_decorators1()
    testing_decorators2()



## THIS IS FILE IS FROM https://github.com/turulomio/reusingcode IF YOU NEED TO UPDATE IT PLEASE MAKE A PULL REQUEST IN THAT PROJECT
## DO NOT UPDATE IT IN YOUR CODE IT WILL BE REPLACED USING FUNCTION IN README

from decimal import Decimal
from logging import warning
from .objects.currency import Currency
from .objects.percentage import Percentage

def valueORempty(v):
    return "" if v is None else v

def list2string(lista):
        """Covierte lista a string"""
        if  len(lista)==0:
            return ""
        if str(lista[0].__class__) in ["<class 'int'>", "<class 'float'>"]:
            resultado=""
            for l in lista:
                resultado=resultado+ str(l) + ", "
            return resultado[:-2]
        elif str(lista[0].__class__) in ["<class 'str'>",]:
            resultado=""
            for l in lista:
                resultado=resultado+ "'" + str(l) + "', "
            return resultado[:-2]

## Reverse function of list2string where class is a str
def string2list_of_strings(s):
    arr=[]
    if s!="":
        arrs=s.split(", ")
        for a in arrs:
            arr.append(a[1:-1])
    return arr

def string2list_of_integers(s, separator=", "):
    """Convers a string of integer separated by comma, into a list of integer"""
    arr=[]
    if s!="":
        arrs=s.split(separator)
        for a in arrs:
            arr.append(int(a))
    return arr
    
    
## Converts a string  to a decimal
def string2decimal(s, type=1):
    if type==1: #2.123,25
        try:
            return Decimal(s.replace(".","").replace(",", "."))
        except:
            return None

## Converts a decimal to a localized number string
def l10nDecimal(dec, digits=2):
    from PyQt5.QtCore import QLocale
    l=QLocale()
    return l.toCurrencyString(float(dec))

## Converts strings True or False to boolean
## @param s String
## @return Boolean
def str2bool(value):
    if value=="0" or value.lower()=="false":
        return False
    elif value=="1" or value.lower()=="true":
        return True

## Converts boolean to  True or False string
## @param s String
## @return Boolean
def bool2string(b):
    if b==True:
        return "VERDADERO"
    return "FALSO"
    
## Function that converts a None value into a Decimal('0')
## @param dec Should be a Decimal value or None
## @return Decimal
def none2decimal0(dec):
    return none2alt(dec,Decimal('0'))

## If a value is None, returns an alternative
def none2alt(value, alternative):
    if value==None:
        return alternative
    return value

## Bytes 2 string
def b2s(b, code='UTF-8'):
    return b.decode(code)
    
def s2b(s, code='UTF8'):
    """String 2 bytes"""
    if s==None:
        return "".encode(code)
    else:
        return s.encode(code)

def c2b(state):
    """QCheckstate to python bool"""
    from PyQt5.QtCore import Qt
    if state==Qt.Checked:
        return True
    else:
        return False

def b2c(booleano):
    """Bool to QCheckstate"""
    from PyQt5.QtCore import Qt
    if booleano==True:
        return Qt.Checked
    else:
        return Qt.Unchecked     

## Returns a list with object in positions removed
def list_remove_positions(l, listindex):
    if l is None:
        warning("I can't remove positions from a None list")
        return None
    r=[]
    for i, o in enumerate(l):
        if i not in listindex:
            r.append(o)
    return r

## LOR is a list of list. Naned List Of Rows, used in myqtablewidget for example
## @param rows LOR
## @param index int with the index of the position where we are going to insert row
## @param column List with the values to add. Must be of the same size of rows
def lor_add_column(rows, index, column):
    if len(rows)!=len(column):
        warning("I can't add a column with different size of LOR")
        return
    r_rows=[]
    for i, row in enumerate(rows):
        r_rows.append(row[0:index] + [column[i],] + row[index:len(row)])
    return r_rows

## LOR is a list of list. Naned List Of Rows, used in myqtablewidget
## @param listindex is a list of column indexes to remove
def lor_remove_columns(rows, listindex):
    r_rows=[]
    for i, row in enumerate(rows):
        r_rows.append(list_remove_positions(row,listindex))
    return r_rows

## LOR is a list of list. Naned List Of Rows, used in myqtablewidget
## @param listindex is a list of column indexes to remove
def lor_remove_rows(rows, listindex):
    return list_remove_positions(rows, listindex) #It's a list but of row

## Return a lor transposed. Changed rows by columns
def lor_transposed(lor):
    r=[]
    columns=len(lor[0])
    for column in range(columns):
        tran_row=[]
        for row in lor:
            tran_row.append(row[column])
        r.append(tran_row)
    return r

## Extract a column from the list of row
def lor_get_row(lor, row):
    return lor[row]

## Extract a column from the list of row
def lor_get_column(lor, column):
    r=[]
    for row in lor:
        r.append(row[column])
    return r

## Return sum of values of a column from and index to and other index position. This method ignores None values
## This method can sum several objects
## @param row
## @param from_index
## @param to_index
## @param zerovalue 0 or Money(self.mem, 0, self.mem.localcurrency)....
def lor_sum_row(row, from_index, to_index, zerovalue=0):
    s=zerovalue
    for i, column in enumerate(row):
        if i>=from_index and i<=to_index:
            if column is not None:
                s=s + column
    return s

## Return sum of values of a column from and index to and other index position. This method ignores None values
## This method can sum several objects
## @zerovalue 0 or Money(self.mem, 0, self.mem.localcurrency)....
def lor_sum_column(lor, column, from_index, to_index, zerovalue=0):
    s=zerovalue
    for i, row in enumerate(lor):
        if i>=from_index and i<=to_index:
            if row[column] is not None:
                s=s + row[column]
    return s

## String to linux shell
#def string2shell(cadena):
#    cadena=str(cadena)
#    cadena=cadena.replace("'","\\'")
#    return cadena

## strint to latex
def string2tex(cadena):
    cadena=str(cadena)
    cadena=cadena.replace('[','$ [ $')
    cadena=cadena.replace(']','$ ] $')
    cadena=cadena.replace('&','\&')
    cadena=cadena.replace('²','$ ^2 $')
    cadena=cadena.replace('#', '\#')
    return cadena

## Converts a string to set inside an XML to a valid XML string
def string2xml(s):
    s=s.replace('"','&apos;' )
    s=s.replace('<','&lt;' )
    s=s.replace('>','&gt;' )
    s=s.replace('&','&amp;' )
    s=s.replace("'",'&apos;' )
    return s

## Converts a string to set inside an XML to a valid XML string
def xml2string(s):
    s=s.replace('&apos;','"')
    s=s.replace('&lt;','<')
    s=s.replace('&gt;','>')
    s=s.replace('&amp;','&')
    s=s.replace('&apos;',"'")
    return s
    
## Converts my common objects to its numeric value
def object2value(o):
    if o.__class__.__name__ in ["int", "float", "Decimal"]:
        return o
    elif o.__class__.__name__ in ["Currency",  "Money", "USD", "EUR"]:
        return o.amount
    elif o.__class__.__name__ == "Percentage":
        return o.value
    return o
        
        
def value2object(value, stringtypes):
    if value is None:
        return None
    if stringtypes=="int":
        return int(value)
    elif stringtypes=="float":
        return float(value)
    elif stringtypes=="Decimal":
        return Decimal(value)
    elif stringtypes==["datetime", "date", "time"]:
        return value
    elif stringtypes in ["EUR", "€"]:
        return Currency(value, "EUR")
    elif stringtypes in ["USD", "$"]:
        return Currency(value, "EUR")
    elif stringtypes=="Percentage":
        return Percentage(value, 1)
    return value

if __name__ == "__main__":
    def print_lor(lor):
        print("")
        for row in lor:
            print(row)

    lor=[]
    column_to_add=[]
    for i in range(10):
        lor.append([1*i,2*i,3*i])
        column_to_add.append(-i)
    print_lor(lor)

    lor=lor_add_column(lor, 0, column_to_add)
    lor=lor_add_column(lor, 2, column_to_add)
    lor=lor_add_column(lor, 5, column_to_add)
    print_lor(lor)

    a=lor_remove_columns(lor,[2,3])
    print_lor(a)
    b=lor_remove_rows(a,[8,9])
    print_lor(b)

    c=lor_transposed(b)
    print_lor(c)



## @namespace officegenerator.odf2xml
## @brief Function to convert a ODF file to xml output
## 
## Executes the following code in bash
## @code
## odf2xml -o "$1.no.xml" "$1"
## cat "$1.no.xml" | xmllint --format -
## @endcode

import argparse
import gettext
import pkg_resources
import platform
import subprocess
import sys

from officegenerator.commons import argparse_epilog,  __version__

try:
    t=gettext.translation('officegenerator',pkg_resources.resource_filename("officegenerator","locale"))
    _=t.gettext
except:
    _=str

def main():
    parser=argparse.ArgumentParser(prog='officegenerator', description=_('Convert a ODF file into an indented xml'), epilog=argparse_epilog(), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--file', action='store', help=_('Odf file to convert'), required=True)
    args=parser.parse_args()

    if platform.system()!="Linux":
        print(_("officegenerator_odf2xml only works on Linux"))
        sys.exit(1)

    p=subprocess.run(["odf2xml", "-o", args.file + ".xml", args.file], stdout=subprocess.PIPE)
    q=subprocess.run(['xmllint', '--format', args.file +".xml"], stdout=subprocess.PIPE, input=p.stdout)
    subprocess.run(['rm', args.file + ".xml"])
    print(q.stdout.decode('UTF-8'))

if __name__ == "__main__":
    main()



## @namespace officegenerator.odf2xml
## @brief Function to convert a ODF file to xml output
## 
## Executes the following code in bash
## @code
## odf2xml -o "$1.no.xml" "$1"
## cat "$1.no.xml" | xmllint --format -
## @endcode

import argparse
import gettext
import pkg_resources
import platform
import subprocess
import sys

from officegenerator.commons import argparse_epilog,  __version__
from os import system
from zipfile import ZipFile

try:
    t=gettext.translation('officegenerator',pkg_resources.resource_filename("officegenerator","locale"))
    _=t.gettext
except:
    _=str

## Generates an xml file with the filename+".xml" name
def generate_xml(filename):
    zipfilename=filename+".zip"
    system("cp '{}' '{}'".format(filename, zipfilename))
    zf = ZipFile(zipfilename)
    for fi in zf.namelist():
        print("### {} ###".format(fi))
        xml=open("tmp.xml", "w")
        print(fi)
        xml.write(zf.read(fi).decode('UTF-8'))
        xml.close()
        q=subprocess.run(['xmllint', '--format', 'tmp.xml'])
        if q.stdout!=None:
            print(q.stdout.decode('UTF-8'))
        subprocess.run(['rm', 'tmp.xml'])
        print("")
    subprocess.run(['rm', zipfilename])

def main():
    parser=argparse.ArgumentParser(prog='officegenerator', description=_('Convert a XLSX file into an indented xml'), epilog=argparse_epilog(), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--file', action='store', help=_('XLSX file to convert'), required=True)
    args=parser.parse_args()

    if platform.system()!="Linux":
        print(_("officegenerator_xlsx2xml only works on Linux"))
        sys.exit(1)

    generate_xml(args.file)

if __name__ == "__main__":
    main()



## @package officegenerator
## @brief Generate office files with predefined styles
from officegenerator.libodfgenerator import ODSStyleColor, ODSStyleCurrency, ColumnWidthODS, ODS_Read, ODS_Write, ODT, ODT_Standard, ODT_Manual_Styles, OdfCell, OdfSheet, guess_ods_style
from officegenerator.libxlsxgenerator import XLSX_Write, XLSX_Read, ColumnWidthXLSX
from officegenerator.commons import columnAdd, rowAdd, Coord, Range, __version__, __versiondate__, __versiondatetime__
from officegenerator.standard_sheets import Model, Model_Auto
from officegenerator.objects.currency import Currency
from officegenerator.objects.percentage import Percentage



from .percentage import Percentage
from .currency import Currency
from decimal import Decimal
class Formula:
    ## Este objeto puede aañdir el valor con string_value and string_type o con value
    ## string_formula debue estar 
    def __init__(self, string_formula=None):
        self.string_formula=string_formula
        
    def __repr__(self):
        return "Formula ({})".format(self.string_formula)
        
    def hasObject(self):
        return hasattr(self, "object")
        
    ## Si no existe _object es porque no se ha establecido el valor. Si existe y es NOne es que es None.
    def setObject(self, object):
        self.object=object
        
    def setObjectFromStrings(self, string_value, string_type):
        self.object=string_value
        

## Returns true if value is a string beginning with = or +
## @param value must be a string
## @return boolean
def isFormula(value):
    if value.__class__.__name__=="str" and len(value)>0 and value[0] in ["=", "+"]:
        return True
    return False

## Generate a formula from a OdfpyCell
def Formula_from_OdfpyCell(odfpycell, string_type=None):
    r=Formula(odfpycell.getAttribute('formula'))
    textvalue=odfpycell.getAttribute('value')
    valuetype=odfpycell.getAttribute('valuetype')
    if odfpycell.getAttribute('valuetype')=="float":
        r.setObject(float(textvalue))
    elif odfpycell.getAttribute('valuetype')=="percentage":
        r.setObject(Percentage(Decimal(textvalue), 1))
    elif odfpycell.getAttribute('valuetype')=="currency":
        print(odfpycell.getAttribute("currency"))
        r.setObject(Currency(Decimal(textvalue), odfpycell.getAttribute("currency")))
    else:
        r.setObject(textvalue)
    print("Creando formula", textvalue, valuetype, "devolviendo", r.object, r.object.__class__)
    return r



## @namespace officegenerator.commons
## @brief Common code to odfpy and openpyxl wrappers
from datetime import datetime
from gettext import translation
from pkg_resources import resource_filename
from logging import info, ERROR, WARNING, INFO, DEBUG, CRITICAL, basicConfig, error, debug
from odf.opendocument import  __version__ as __odfpy_version__
from os import path
from shutil import copyfile
from subprocess import run, PIPE
from tempfile import TemporaryDirectory

__version__ = '1.28.0'
__versiondatetime__=datetime(2021, 2, 27, 16, 11)
__versiondate__=__versiondatetime__.date()

try:
    t=translation('officegenerator',resource_filename("officegenerator","locale"))
    _=t.gettext
except:
    _=str

## Function used in argparse_epilog
## @return String
def argparse_epilog():
    return _("Developed by Mariano Muñoz 2015-{}").format(__versiondate__.year)


## Allows to operate with columns letter names
## @param letter String with the column name. For example A or AA...
## @param number Columns to move
## @return String With the name of the column after movement
def columnAdd(letter, number):
    letter_value=column2number(letter)+number
    return number2column(letter_value)


def rowAdd(letter,number):
    return str(int(letter)+number)

## Convierte un número  con el numero de columna al nombre de la columna de hoja de datos
##
## Number to Excel-style column name, e.g., 1 = A, 26 = Z, 27 = AA, 703 = AAA.
def number2column(n):
    name = ''
    while n > 0:
        n, r = divmod (n - 1, 26)
        name = chr(r + ord('A')) + name
    return name

## Convierte una columna de hoja de datos a un número
##
## Excel-style column name to number, e.g., A = 1, Z = 26, AA = 27, AAA = 703.
def column2number(name):
    n = 0
    for c in name:
        n = n * 26 + 1 + ord(c) - ord('A')
    return n

## Converts a column name to a index position (number of column -1)
def column2index(name):
    return column2number(name)-1

## Convierte el nombre de la fila de la hoja de datos a un índice, es decir el número de la fila -1
def row2index(number):
    return int(number)-1

## Covierte el nombre de la fila de la hoja de datos a un  numero entero que corresponde con el numero de la fila
def row2number(strnumber):
    return int(strnumber)

## Convierte el numero de la fila al nombre de la fila en la hoja de datos , que corresponde con un string del numero de la fila
def number2row(number):
    return str(number)
    
## Convierte el indice de la fila al numero cadena de la hoja de datos
def index2row(index):
    return str(index+1)
    
## Convierte el indice de la columna a la cadena de letras de la columna de la hoja de datos
def index2column(index):
    return number2column(index+1)

def ODFPYversion():
    return __odfpy_version__.split("/")[1]

## Class that manage spreadsheet coordinates (letter + number)
class Coord:
    def __init__(self, strcoord):
        self.letter, self.number=self.__extract(strcoord)
    def __repr__(self):
        return "Coord <{}>".format(self.string())

    def __eq__(self, b):
        b=Coord.assertCoord(b)
        if self.letter==b.letter and self.number==b.number:
            return True
        return False
        
    def __extract(self, strcoord):
        if strcoord.find(":")!=-1:
            print("I can't manage range coord")
            return
        letter=""
        number=""
        for l in strcoord:
            if l.isdigit()==False:
                letter=letter+l
            else:
                number=number+l
        return (letter,number)

    ## Returns Coord string
    ## @return string For example "Z1"
    def string(self):
        return self.letter+self.number

    ## Add a number of rows to the Coord and return itself
    ## @param num Integer Can be positive and negative. When num is negative, if Coord.number is less than 1, returns 1
    def addRow(self, num=1):
        if self.numberIndex()+num<0:
            self.number="1"
        else:
            self.number=rowAdd(self.number, num)
        return self

    ## Add a number of columns/letters to the Coord and return itself
    ## @param num Integer. Can be positive and negative. When num is negative, if Coord.letter is less than A, returns A.
    def addColumn(self, num=1):
        if self.letterIndex()+num<0:
            self.letter="A"
        else:
            self.letter=columnAdd(self.letter, num)
        return self

    ## Add a number of rows to the Coord and return a copy of the object
    ## @param num Integer Can be positive and negative. When num is negative, if Coord.number is less than 1, returns 1
    def addRowCopy(self, num=1):
        r=Coord(self.string())
        if r.numberIndex()+num<0:
            r.number="1"
        else:
            r.number=rowAdd(r.number, num)
        return r

    ## Add a number of columns/letters to the Coord and return a copy of the objject
    ## @param num Integer. Can be positive and negative. When num is negative, if Coord.letter is less than A, returns A.
    def addColumnCopy(self, num=1):
        r=Coord(self.string())
        if r.letterIndex()+num<0:
            r.letter="A"
        else:
            r.letter=columnAdd(r.letter, num)
        return r

    ## Reterns the letter/column index (letterPosition()-1)
    ## @return int
    def letterIndex(self):
        return column2index(self.letter)

    ## Returns the letter/column position
    ## @return int 
    def letterPosition(self):
        return column2number(self.letter)

    ## Returns the number/row index (numberPosition()-1)
    def numberIndex(self):
        return row2index(self.number)

    ## Returns the number/row position
    def numberPosition(self):
        return row2number(self.number)

    @staticmethod
    def assertCoord(o):
        if o.__class__==Coord:
            return o
        elif o.__class__==str:
            return Coord(o)
        else:
            error("{} is not a coord".format(o))
            



## Class that manages spreadsheet Ranges for ods and xlsx
class Range:
    def __init__(self,strrange):
        self.start, self.end=self.__extract(strrange)
        
    def __repr__(self):
        return "Range <{}>".format(self.string())

    ##Return the outcome of the test b in a. Note the reversed operands.
    def __contains__(self, b):
        if (    b.letterIndex()>=self.start.letterIndex() and 
                b.letterIndex()<=self.end.letterIndex() and 
                b.numberIndex()>=self.start.numberIndex() and 
                b.numberIndex()<=self.end.numberIndex())==True:
            return True
        return False

    ## Returns a list of rows of all Coord objects in the range
    def coords(self):
        r=[]
        for row in range(self.numRows()):
            tmprow=[]
            for column in range(self.numColumns()):
                tmprow.append(self.start.addRowCopy(row).addColumnCopy(column))
            r.append(tmprow)
        return r

    ## Converts a string to a Range. Returns None if conversion can't be done
    def __extract(self,range):
        if range.find(":")==-1:
            print("I can't manage this range")
            return
        a=range.split(":")
        return (Coord(a[0]), Coord(a[1]))

    ## String of a range in spreadsheets
    def string(self):
        return "{}:{}".format(self.start.string(), self.end.string())

    ## Number of range of the range
    def numRows(self):
        return row2number(self.end.number)-row2number(self.start.number) +1

    ## Number of columns of the range
    def numColumns(self):
        return column2number(self.end.letter)-column2number(self.start.letter) +1

    ## Returns a Range object even o is a str or a Range
    @staticmethod
    def assertRange(o):
        if o.__class__==Range:
            return o
        elif o.__class__==str:
            return Range(o)


    ## Adds a row to the end Coord, so it adds a row to the range
    def addRowAfter(self, num=1):
        self.end=self.end.addRow(num)
        return self

    ## Adds a column to the end Coord, so it adds a column to the range
    def addColumnAfter(self, num=1):
        self.end=self.end.addColumn(num)
        return self

    ## Adds a row to the top of the start Coord, so it adds a row to the range. If start Coord number is 1 returns the same Coord
    def addRowBefore(self, num=1):
        self.start=self.start.addRow(-num)
        return self

    ## Adds a column to the end Coord, so it adds a column to the range. If start Coord letter is A, returns the same Coord
    def addColumnBefore(self, num=1):
        self.start=self.start.addColumn(-num)
        return self

## Sets debug sustem, needs
## @param args It's the result of a argparse     args=parser.parse_args()        
def addDebugSystem(level):
    logFormat = "%(asctime)s.%(msecs)03d %(levelname)s %(message)s [%(module)s:%(lineno)d]"
    dateFormat='%F %I:%M:%S'

    if level=="DEBUG":#Show detailed information that can help with program diagnosis and troubleshooting. CODE MARKS
        basicConfig(level=DEBUG, format=logFormat, datefmt=dateFormat)
    elif level=="INFO":#Everything is running as expected without any problem. TIME BENCHMARCKS
        basicConfig(level=INFO, format=logFormat, datefmt=dateFormat)
    elif level=="WARNING":#The program continues running, but something unexpected happened, which may lead to some problem down the road. THINGS TO DO
        basicConfig(level=WARNING, format=logFormat, datefmt=dateFormat)
    elif level=="ERROR":#The program fails to perform a certain function due to a bug.  SOMETHING BAD LOGIC
        basicConfig(level=ERROR, format=logFormat, datefmt=dateFormat)
    elif level=="CRITICAL":#The program encounters a serious error and may stop running. ERRORS
        basicConfig(level=CRITICAL, format=logFormat, datefmt=dateFormat)
    info("Debug level set to {}".format(level))

## When top left cell is None in freezeAndSelect, both ods and xlsx returns a default topLeftCell
def topLeftCellNone(freeze_coord, selected_coord):
        freeze_coord=Coord.assertCoord(freeze_coord)
        selected_coord=Coord.assertCoord(selected_coord)
        #Get the select coord - 20 rows and - 10 columns
        minus_coord=selected_coord.addRowCopy(-20).addColumnCopy(-10)
        if minus_coord.letterIndex()<=freeze_coord.letterIndex():#letter
            letter=freeze_coord.letter
        else:
            letter=minus_coord.letter
        if minus_coord.numberIndex()<=freeze_coord.numberIndex():#number
            number=freeze_coord.number
        else:
            number=minus_coord.number
        return Coord(letter+number)

## Creates a Coord object from spreadsheet letters
def Coord_from_letters(column, letter):
    return Coord(column+letter)

## Creates a Coord object from spreadsheet index coords
def Coord_from_index(column_index, row_index):
    return Coord(index2column(column_index)+index2row(row_index))


### This command has problems with concurrency seems that three is a hidden lock. Do not use with concurrency
#def convert_command(filename, ouput_dir, to_format):
#    command="loffice --nolockcheck --headless --convert-to {} --outdir '{}' '{}'".format(to_format, ouput_dir, filename)
#    r=run(command, shell=True, stdout=PIPE)
#    if r.returncode!=0:
#        print ("Error with command: {}".format(command))
#        return None
        
        
        
## This command has problems with concurrency seems that three is a hidden lock. Do not use with concurrency
def convert_command(filename_from, filename_to):
    filename_from_name, filename_from_extension = path.splitext(path.basename(filename_from))    
    filename_to_name, filename_to_extension = path.splitext(path.basename(filename_to))    
    with TemporaryDirectory(prefix="officegenerator_") as tmp_name:
        command="loffice --nolockcheck --headless --convert-to {} --outdir '{}' '{}'".format(filename_to_extension[1:], tmp_name, filename_from)
        r=run(command, shell=True, stdout=PIPE)
        if r.returncode!=0:
            error("Error with command: {}".format(command))
            return None            
        tmp_converted_filename="{}/{}{}".format(tmp_name, filename_from_name, filename_to_extension)
        debug("Libreoffice conversion from '{}' to '{}'".format(filename_from, filename_to))
        copyfile(tmp_converted_filename, filename_to)
    return filename_to

def generate_formula_total_string(key, coord_from, coord_to):
    if key == "#SUM":
        s="=SUM({}:{})".format(coord_from.string(), coord_to.string())
    elif key == "#AVG":
        s="=AVERAGE({}:{})".format(coord_from.string(), coord_to.string())
    elif key == "#MEDIAN":
        s="=MEDIAN({}:{})".format(coord_from.string(), coord_to.string())
    else:
        s=key
    return s



from logging import warning
from officegenerator.commons import Coord, index2column, index2row, columnAdd
from officegenerator.casts import lor_remove_columns, lor_remove_rows,  list_remove_positions, lor_add_column, lor_get_column, lor_get_row
from officegenerator.libodfgenerator import guess_ods_style, ODS_Write
from officegenerator.libxlsxgenerator import XLSX_Write
from officegenerator.objects.currency import currency_symbol

## Models are used to generate standard sheets
## We will use them to generate fast sheets with or without totals
## See the __main__ for examples

class ModelStyles:
    hh=0# Only horizontal header
    hv=1
    hht=2 #Only horizontal header with last row of totals
    hhv=3 # Only vertical header with last column of totals
    hhthhv=4 # Horizontal and vertical headers with last row and column of totals

class Model:
    def __init__(self):
        self.hh=None
        self.vh=None
        self.ht_definition=None
        self.vt_definition=None

    def setTitle(self, title):
        self.title=title
        
    ## If you want only vertical headers you must add self.setHorizontalHeaders(None, sizes=10)
    ## @param hh List or NOne
    ## @param sizes int or List in cm, to generate document use, columnSizes_for_ods
    def setHorizontalHeaders(self, hh, sizes=10):
        self.hh=hh
        if hh==None:#Only vertical headers
            columns=100
        else:
            columns=len(self.hh)+1
        if sizes.__class__.__name__=="int":
            self.hh_sizes=[sizes]*columns
        else:
            self.hh_sizes=sizes

    ## @param hh List or NOne
    ## @param size int with the witdth in cm of the vertical header
    def setVerticalHeaders(self, vh, size=5):
        self.vh=vh
        self.vh_size=size

#    ## REturns if a row is operable. It's used to generate automatic operable models
#    def isRowOperable(self, row_index, operable_from):
#        r=True
#        for cell in self.data[row_index]:
#            print( operable_from, row_index, cell,   cell.__class__.__name__)
#            if cell.__class__.__name__ in ["str",  "datetime", "date"]:
#                r=False
#                break
#        return r
#            
#    ## REturns if a row is operable. It's used to generate automatic operable models
#    def isColumnOperable(self, column_index, operable_from):
#        r=True
#        for i, row in enumerate(self.data):
#            if i>=operable_from and row[column_index].__class__.__name__ in ["str",  "datetime", "date" ]:
#                r= False
#                break
#        return r
        

    ## Order data columns. None values are set at the beginning
    def order_with_none(self, column_index, reverse=False, none_at_top=True):
        nonull=[]
        null=[]
        for row in self.data:
            if row[column_index] is None:
                null.append(row)
            else:
                nonull.append(row)
        nonull=sorted(nonull, key=lambda c: c[column_index], reverse=reverse)
        if none_at_top==True:#Set None at top of the list
            self.data= null+nonull
        else:
            self.data=nonull+null

    ## Converts self.hh_sizes and self.hv_size in cm to ods sizes. It returns them together
    def columnSizes_for_ods(self):
        r=[]
        factor=30
        if self.vh is not None:
            r.append(self.vh_size*factor)
        for arg in self.hh_sizes:
                r.append(arg*factor)
        return r
        
    ## Returns the number of rows in data
    def numDataRows(self):
        return len(self.data)

    ## Returns the number of rows in data
    def numDataColumns(self):
        return len(self.data[0])
        
    ## Converts self.hh_sizes in cm to xlsx sizes
    def columnSizes_for_xlsx(self):
        r=[]
        factor=6
        if self.vh is not None:
            r.append(self.vh_size*factor)
        for arg in self.hh_sizes:
                r.append(arg*factor)
        return r
        
    ## Converts self.hh_sizes in cm to odt sizes,proporcional to the parameter
    def columnSizes_for_odt(self, tablesize):
        cm=[]
        factor=1
        if self.vh is not None:
            cm.append(self.vh_size*factor)
        for arg in self.hh_sizes:
                cm.append(arg*factor)
        #Until here are cm but can oversize the maximum
        r=[]
        sum_=sum(cm)
        for size in cm:
            r.append(tablesize*size/sum_)
        return r

    def setData(self, data):
        self.data=data
        
    ## Used to remove Columns in the self.data
    ## @param columnList List of integers with the column index to remove
    def removeColumns(self, columnList):
        if self.hh is not None:
            self.hh=list_remove_positions(self.hh, columnList)
        self.hh_sizes=list_remove_positions(self.hh_sizes, columnList)
        if self.data is not None:
            self.data=lor_remove_columns(self.data, columnList)
        else:
            warning("I can't remove columns if self.data is None")

    ## Used to remove Columns in the self.data
    ## @param columnList List of integers with the column index to remove
    def removeRows(self, rowList):
        if self.vh is not None:
            self.vh=list_remove_positions(self.vh, rowList)
        if self.data is not None:
            self.data=lor_remove_rows(self.data, rowList)
        else:
            warning("I can't remove rows if self.data is None")

    ## @param title String with the title of the sheet
    ## @param columns_title List of Strings
    ## @param data list of list
    ## @param sizes List of integers
    def xlsx_sheet(self, doc):
        doc.createSheet(self.title)
        doc.setColumnsWidth(self.columnSizes_for_xlsx())
        if self.hh is not None:
            doc.overwrite(self.__getFirstContentCoord().addRow(-1), [self.hh], doc.stOrange)
        if self.vh is not None:
            for i, header in enumerate(self.vh):
                doc.overwrite(self.__getFirstContentCoord().addColumn(-1).addRow(i), header, doc.stGreen, alignment="left")
        if self.__mustFillA1()==True:
            doc.overwrite("A1", " ",  doc.stOrange)

        if self.numDataRows()>0:
            for number, row in enumerate(self.data):
                for letter,  field in enumerate(row):
                    doc.overwrite(self.__getFirstContentCoord().addRow(number).addColumn(letter), field, style=doc.stWhite)

            #Fills horizontal  total
            if self.ht_definition is not None:
                for letter, definition in enumerate(self.ht_definition):
                    class_=self.__object_to_formula_classname(self.data[0][letter])
                    if self.__is_total_key(definition):
                        doc.overwrite_formula(self.__getFirstContentCoord().addRow(self.numDataRows()).addColumn(letter), self.__calculate_horizontal_total("xlsx", letter), class_,  style=doc.stGrayLight)
                    else:
                        doc.overwrite(self.__getFirstContentCoord().addRow(self.numDataRows()).addColumn(letter), self.__calculate_horizontal_total("xlsx", letter), style=doc.stGrayLight)
            #Fills vertical total
            if self.vt_definition is not None:
                for number, definition in enumerate(self.vt_definition):
                    class_=self.__object_to_formula_classname(self.data[0][self.numDataColumns()-1])
                    value=self.__calculate_vertical_total("xlsx", number)
                    if self.__is_total_key(definition):
                        doc.overwrite_formula(Coord("A1").addRow(number).addColumn(self.numDataColumns()), value, class_,  style=doc.stGrayLight)
                    else:
                        doc.overwrite(Coord("A1").addRow(number).addColumn(self.numDataColumns()), value, style=doc.stGrayLight)
        self.__setFreezeAndSelect(doc)

    ## Function neeeded to change formula types, due to is a string but needs to be changed to currency symbol. to use __setFormulaNumberFormat
    def __object_to_formula_classname(self, o):
        if o.__class__.__name__ in ("Currency", "Money"):
            return currency_symbol(o.currency) 
        else:
            return o.__class__.__name__

    ## @param title String with the title of the sheet
    ## @param columns_title List of Strings
    ## @param data list of list
    ## @param sizes List of integers
    def ods_sheet(self, doc):
        s=doc.createSheet(self.title)
        s.setColumnsWidth(self.columnSizes_for_ods())
        if self.hh is not None:
            s.add(self.__getFirstContentCoord().addRow(-1), [self.hh],  "OrangeCenter")
        if self.vh is not None :
            for i, header in enumerate(self.vh):
                s.add(self.__getFirstContentCoord().addColumn(-1).addRow(i), header, "GreenLeft")
        if self.__mustFillA1()==True:
            s.add("A1", "", "OrangeCenter")

        if self.numDataRows()>0: #Only must be executed with data
            for number, row in enumerate(self.data):
                for letter,  field in enumerate(row):
                    s.add(self.__getFirstContentCoord().addRow(number).addColumn(letter), field)
                    
            #Fills horizontal 
            if self.ht_definition is not None:
                for letter, definition in enumerate(self.ht_definition):
                    value= self.__calculate_horizontal_total("ods", letter)
                    style=guess_ods_style("GrayLight", self.data[self.ht_index_from][letter])
                    s.add(Coord("A1").addRow(self.numDataRows()+1).addColumn(letter), value, style )
            #Fills vertical
            if self.vt_definition is not None:
                for number, definition in enumerate(self.vt_definition):
                    value=self.__calculate_vertical_total("ods", number)
                    style=guess_ods_style("GrayLight", self.data[self.vt_index_from][letter])
                    s.add(Coord("A1").addRow(number).addColumn(self.numDataColumns()), value,  style)

        self.__setFreezeAndSelect(s)
        return s

    ## @param type can be "ods","xlsx","odt","value"
    ## See setHorizontalTotalDefinition doc for available keys
    def __calculate_horizontal_total(self, type, column_index):
        key=self.ht_definition[column_index]
        column=index2column(column_index)
        total_from=index2row(self.ht_index_from)
        r=key
        if type in("ods",  "xlsx"):
            if key=="#SUM":
                r= "=SUM({0}{1}:{0}{2})".format(column, total_from, self.numDataRows()+1)
            elif key=="#AVG":
                r= "=AVERAGE({0}{1}:{0}{2})".format(column, total_from, self.numDataRows()+1)
            elif key=="#MEDIAN":
                r= "=MEDIAN({0}{1}:{0}{2})".format(column, total_from, self.numDataRows()+1)
        elif type=="value":
            if key=="#SUM":
                r= sum(lor_get_column(self.data, column))
        return r

    ## @param type can be "ods","xlsx","odt","value"
    ## See setHorizontalTotalDefinition doc for available keys
    def __calculate_vertical_total(self, type, row_index):
        key=self.vt_definition[row_index]
        row=index2row(row_index)
        total_from=index2column(self.vt_index_from)
        r=key
        if type in("ods",  "xlsx"):
            if key=="#SUM":
                r= "=SUM({0}{1}:{2}{1})".format(total_from, row, columnAdd(total_from, self.numDataColumns()-self.vt_index_from-1))
            elif key=="#AVG":
                r= "=AVERAGE({0}{1}:{2}{1})".format(total_from, row, columnAdd(total_from, self.numDataColumns()-self.vt_index_from-1))
            elif key=="#MEDIAN":
                r= "=MEDIAN({0}{1}:{2}{1})".format(total_from, row, columnAdd(total_from, self.numDataColumns()-self.vt_index_from-1))
        elif type=="value":
            if key=="#SUM":
                r= sum(lor_get_row(self.data, row))
        return r
        
        
    def __is_total_key(self, s):
        if s in ["#SUM","#AVG","#MEDIAN"]:
            return True
        return False

    ## Generates a odt table object from model
    ## @param doc odt document
    ## @param tablesize float in cm where the table is going to be placed in the paper(document)
    ## @param fontsiz1e int withe the size of the font in the document
    ## @param after officegenerator after parameter
    def odt_table(self, doc, tablesize, fontsize, after=True):
        if self.vh is not None:
            data=lor_add_column(self.data, 0, self.vh)
        else:
            data=self.data
        if self.__mustFillA1()==True:
            hh=[" ", ] + self.hh
        else:
            hh=self.hh
        return doc.table(hh, data, self.columnSizes_for_odt(tablesize), fontsize, self.title, after)

    def __getFirstContentCoord(self):
        #firstcontentletter and firstcontentnumber
        if self.hh is None and self.vh is not None:
            return Coord("B1")
        elif self.hh is not None and self.vh is None:
            return Coord("A2")
        elif self.hh is not None and self.vh is not None:
            return Coord("B2")
        elif self.hh is None and self.vh is None:
            return Coord("A1")
        
    ## @param ref Reference to object: ods: sheet, xlsx: doc
    def __setFreezeAndSelect(self, ref):
        if self.ht_definition is not None:
            ref.freezeAndSelect(self.__getFirstContentCoord(), self.__getFirstContentCoord().addRow(self.numDataRows()).addColumn(self.numDataColumns()-1))
        elif self.vt_definition is not None:
            ref.freezeAndSelect(self.__getFirstContentCoord(), self.__getFirstContentCoord().addRow(self.numDataRows()-1).addColumn(self.numDataColumns()))
        else:
            ref.freezeAndSelect(self.__getFirstContentCoord(),self.__getFirstContentCoord().addRow(self.numDataRows()-1).addColumn(self.numDataColumns()-1))
            
        
    ## Return if A1 must be filled for a better view
    ## @param bool
    def __mustFillA1(self):
        if self.hh is not None and self.vh is not None:
            return True
        return False
        
    
    ## Available keys:
    ## - #SUM
    ## - #AVG
    ## - #COUNT
    ## - #MEDIAN
    ## @param definition_list List with strings and keys
    ## @param totals_index_from Column index from with totals are generated
    def setHorizontalTotalDefinition(self, definition_list, totals_index_from=1):
        self.ht_definition=definition_list    ## Available keys:
        self.ht_index_from=totals_index_from

    ## See setHorizontalTotalDefinition doc for available keys
    ## @param definition_list List with strings and keys
    ## @param totals_index_from Column index from with totals are generated
    def setVerticalTotalDefinition(self, definition_list, totals_index_from=1):
        self.vt_definition=definition_list
        self.vt_index_from=totals_index_from
        
    ## Generates an xlsx file from model
    def xlsx_file(self, filename):
        xlsx=XLSX_Write(filename)
        self.xlsx_sheet(xlsx)
        xlsx.remove_sheet_by_id(0)
        xlsx.save()

    ## Generates an ods file from model
    def ods_file(self, filename):
        ods=ODS_Write(filename)
        self.ods_sheet(ods)
        ods.save()
        
        
## Can show ht and vt. from an index
## @param title Title of the model and sheet
## @param hh List with Horizontal headers
## @param data lor with model data
## @param allkeys Operation key for all fields
## @param row_index_from Defines if some rows are not going to be operated in formulas
## @param column_index_from Defines if some columns are not going to be operated in formulas
## @param ht Boolean to show horizontal totals
## @param vt Boolean to show vertical totals
## @param skip_ht List of index column to skip some totals
## @param skip_vt List of index rows to skip some totals
def Model_Auto(title, hh, data, allkeys="#SUM", sizes=5, row_index_from=1, column_index_from=1, ht=True, vt=True, skip_ht=[],skip_vt=[]):
    m=Model()
    m.setTitle(title)
    m.setHorizontalHeaders(hh, sizes)
    m.setData(data)

    if ht is True:
        hdef=["Total", ]
        hdef=hdef+[""]*(column_index_from-1)# Empty totals
        for index in range(column_index_from, m.numDataColumns()):
            if index in skip_ht:
                hdef=hdef+[""]
            else:
                hdef=hdef+[allkeys]
        m.setHorizontalTotalDefinition(hdef, totals_index_from=column_index_from)

    if vt is True:
        vdef=["Total", ]
        vdef=vdef+[""]*(row_index_from-1)# Empty totals
        for index in range(row_index_from, m.numDataRows()+1):
            if index in skip_vt:
                vdef=vdef+[""]
            else:
                vdef=vdef+[allkeys]
        
        if ht is True:
            vdef=vdef + [allkeys]#Crossed total
            
        m.setVerticalTotalDefinition(vdef, totals_index_from=column_index_from)
    return m



## THIS IS FILE IS FROM https://github.com/turulomio/reusingcode IF YOU NEED TO UPDATE IT PLEASE MAKE A PULL REQUEST IN THAT PROJECT
## DO NOT UPDATE IT IN YOUR CODE IT WILL BE REPLACED USING FUNCTION IN README


from datetime import datetime, timezone
from urllib.request import urlopen
from json import loads
from os import  path

## Get Github file modification datetime
## https://api.github.com/repos/turulomio/xulpymoney/commits?path=products.xlsx
## @return datetime or None if can't find it
def get_file_modification_dtaware(user,project,path):
    try:
        url="https://api.github.com/repos/{}/{}/commits?path={}".format(user,project,path)
        print(url)
        bytes_j = urlopen(url).read()
        j=loads(bytes_j.decode('UTF-8'))
        dtnaive= datetime.strptime(j[0]['commit']['author']['date'], "%Y-%m-%dT%H:%M:%SZ")
        return dtnaive.replace(tzinfo=timezone.utc)
    except:
        None

## Downloads file to a destiny directory
def download_from_github(user,repository,path_filename, destiny_directory):
    destiny_path='{}/{}'.format(destiny_directory,path.basename(path_filename))
    download_from_github_to_path(user, repository, path_filename, destiny_path)

## Downloads file to a new file path
def download_from_github_to_path(user,repository,path_filename, destiny_path):
    command="https://raw.githubusercontent.com/{}/{}/master/{}".format(user,repository, path_filename)
    print(command)
    response = urlopen(command, timeout = 5)
    content = response.read()
    f = open(destiny_path, 'wb' )
    f.write( content )
    f.close()
    print("Updating {} from https://github.com/turulomio/reusingcode/{}".format(destiny_path,path_filename))

## Get the latest release tag of code releases in a repository
def get_latest_release(user, repository):
    url="https://api.github.com/repos/{}/{}/releases/latest".format(user,repository)
    try:
        b_json=urlopen(url).read()
        j=loads(b_json.decode('UTF-8'))
        return j['tag_name']
    except:
        return None

if __name__ == '__main__':
    print(get_file_modification_dtaware("turulomio","xulpymoney","products.xlsx"))
    download_from_github("turulomio", "xulpymoney", "doc/xulpymoney.odt", "./")
    download_from_github_to_path("turulomio", "xulpymoney", "doc/xulpymoney.odt", "xulpymoney.odt")
    print("Latest xulpymoney release:", get_latest_release("turulomio","xulpymoney"))
    print("Latest caloriestracker release:", get_latest_release("turulomio","caloriestracker"))
    print("Latest reusingcode release:", get_latest_release("turulomio","reusingcode"))



## @namespace officegenerator.libodfgenerator
## @brief Package that allows to read and write Libreoffice ods and odt files
## This file is from the Xulpymoney project. If you want to change it. Ask for project administrator

## odf Element and Text inherits from Node and inherits from xml.Dom.Node https://docs.python.org/3.5/library/xml.dom.html
## Text can't have children

from datetime import datetime, time
import gettext
from decimal import Decimal
from logging import info, debug
from odf.opendocument import OpenDocumentSpreadsheet,  OpenDocumentText,  load
from odf.style import Footer, FooterStyle, HeaderFooterProperties, Style, TextProperties, TableColumnProperties, Map,  TableProperties,  TableCellProperties, PageLayout, PageLayoutProperties, ParagraphProperties,  ListLevelProperties,  MasterPage
from odf.number import  TimeStyle, CurrencyStyle, CurrencySymbol,  Number, NumberStyle, Text,  PercentageStyle,  DateStyle, Year, Month, Day, Hours, Minutes, Seconds, Boolean, BooleanStyle
from odf.text import P,  H,  Span, ListStyle,  ListLevelStyleBullet,  List,  ListItem, ListLevelStyleNumber,  OutlineLevelStyle,  OutlineStyle,  PageNumber,  PageCount
from odf.table import Table, TableColumn, TableRow, TableCell,  TableHeaderRows
from odf.draw import Frame, Image
from odf.dc import Creator, Description, Title, Date, Subject
from odf.meta import InitialCreator, Keyword, CreationDate
import odf.element

from odf.config import ConfigItem, ConfigItemMapEntry, ConfigItemMapIndexed, ConfigItemMapNamed,  ConfigItemSet
from odf.office import Annotation
from officegenerator.commons import number2column,  number2row,  Coord, Range, topLeftCellNone, column2index, Coord_from_index, row2index, convert_command, generate_formula_total_string
from officegenerator.objects.currency import Currency
from officegenerator.objects.formula import Formula_from_OdfpyCell, isFormula, Formula
from officegenerator.datetime_functions import dtnaive2string
from officegenerator.objects.percentage import Percentage
from os import path, remove, makedirs, sep
from pkg_resources import resource_filename
from tempfile import TemporaryDirectory

try:
    t=gettext.translation('officegenerator', resource_filename("officegenerator","locale"))
    _=t.gettext
except:
    _=str

## Type class that defines predefined width columns
class ColumnWidthODS:
    Date=60
    Datetime=100
    XS=12
    S=25
    M=50
    L=100
    XL=150
    XXL=200
    XXXL=250
    XXXXL=300

class ODS_Read:
    def __init__(self, filename):
        self.doc=load(filename)#doc is only used in this function. All is generated in self.doc
        self.filename=filename

    def getSheetElementByName(self, name):
        """
            Devuelve el elemento de sheet buscando por su nombre
        """
        for numrow, sheet in  enumerate(self.doc.spreadsheet.getElementsByType(Table)):
            if sheet.getAttribute("name")==name:
                return sheet
        return None        

    def getSheetElementByIndex(self, index):
        """
            Devuelve el elemento de sheet buscando por su posición en el documento
        """
        try:
            return self.doc.spreadsheet.getElementsByType(Table)[index]
        except:
            return None
#        
#    ## Returns a list of rows with the odfcells of the sheet
#    ## @param sheet_index Integer index of the sheet
#    ## @param skip_up int. Number of rows to skip at the begining of the list of rows (lor)
#    ## @param skip_down int. Number of rows to skip at the end of the list of rows (lor)
#    ## @return Returns a list of rows of object values
#    def cells(self, sheet_index, skip_up=0, skip_down=0):
#        sheet_element=self.getSheetElementByIndex(sheet_index)        
#        rows=sheet_element.getElementsByType(TableRow) #Uses ODFPY cell to boost performance
#        r=[]
#        for row in rows[skip_up:len(rows)-skip_down]:
#            tmprow=[]
#            for cell in row.getElementsByType(TableCell):
#                tmprow.append(self._getCellValue_from_odfpy_cell(cell))
#            r.append(tmprow)
#        return r

        
#    ## REturns a lor of odfpy cells
#    def cellsOdfPy(self, sheet_index):
#        r=[]
#        sheet_element=self.getSheetElementByIndex(sheet_index)
#        for row in sheet_element.getElementsByType(TableRow):
#            r_row=[]
#            for cell in row.getElementsByType(TableCell):
#                r_row.append(cell)
#            r.append(r_row)
#        return r

    ## Returns a list of rows with the values of the sheet
    ## @param sheet_index Integer index of the sheet
    ## @param skip_up int. Number of rows to skip at the begining of the list of rows (lor)
    ## @param skip_down int. Number of rows to skip at the end of the list of rows (lor)
    ## @return Returns a list of rows of object values
    def values(self, sheet_index, skip_up=0, skip_down=0):
        sheet_element=self.getSheetElementByIndex(sheet_index)        
        rows=sheet_element.getElementsByType(TableRow) #Uses ODFPY cell to boost performance
        r=[]
        for row in rows[skip_up:len(rows)-skip_down]:
            tmprow=[]
            for cell in row.getElementsByType(TableCell):
                tmprow.append(self._getCellValue_from_odfpy_cell(cell))
            r.append(tmprow)
        return r

    ## @param sheet_index Integer index of the sheet
    ## @param range_ Range object to get values. If None returns all values from sheet
    ## @return Returns a list of rows of object values
    def values_by_range(self, sheet_index, range_):
        sheet_element=self.getSheetElementByIndex(sheet_index)        
        rows=sheet_element.getElementsByType(TableRow) #Uses ODFPY cell to boost performance
        r=[]
        range_=Range.assertRange(range_)
        for number_index, row in enumerate(rows):
            tmprow=[]
            for letter_index, cell in enumerate(row.getElementsByType(TableCell)):
                if Coord_from_index(letter_index, number_index) in range_:
                    tmprow.append(self._getCellValue_from_odfpy_cell(cell))
            r.append(tmprow)
        return r
    
    ## @param sheet_index Integer index of the sheet
    ## @param column_letter Letter of the column to get values
    ## @param skip Integer Number of top rows to skip in the result
    ## @return List of values
    def getColumnValues(self, sheet_index, column_letter, skip_up=0, skip_down=0):
        r=[]
        sheet_element=self.getSheetElementByIndex(sheet_index)        
        rows=sheet_element.getElementsByType(TableRow) #Uses ODFPY cell to boost performance
        for row in rows[skip_up:len(rows)-skip_down]:
            cell=row.getElementsByType(TableCell)[column2index(column_letter)]
            r.append(self._getCellValue_from_odfpy_cell(cell))
        return r    

    ## @param sheet_index Integer index of the sheet
    ## @param row_number String Number of the row to get values
    ## @param skip Integer Number of top rows to skip in the result
    ## @return List of values
    def getRowValues(self, sheet_index, row_number, skip_left=0, skip_right=0):
        r=[]
        sheet_element=self.getSheetElementByIndex(sheet_index)        
        #Uses ODFPY cell to boost performance
        row=sheet_element.getElementsByType(TableRow)[row2index(row_number)]
        cells=row.getElementsByType(TableCell)
        for cell in cells[skip_left:len(cells)-skip_right]:
            r.append(self._getCellValue_from_odfpy_cell(cell))
        return r    

    ## Return a Range object with the limits of the index sheet
    def getSheetRange(self, sheet_index):
        endcoord=Coord("A1").addRow(self.rowNumber(sheet_index)-1).addColumn(self.columnNumber(sheet_index)-1)
        return Range("A1:"+endcoord.string())
        
    def rowNumber(self, sheet_index):
        """
            Devuelve el numero de filas de un determinado sheet_element
        """
        sheet_element=self.getSheetElementByIndex(sheet_index)
        return len(sheet_element.getElementsByType(TableRow))
        
    def columnNumber(self, sheet_index):
        """
            Devuelve el numero de filas de un determinado sheet_element
        """
        sheet_element=self.getSheetElementByIndex(sheet_index)
        return len(sheet_element.getElementsByType(TableColumn))
        
    def getOdfPyCell(self, sheet_index, coord):
        coord=Coord.assertCoord(coord)
        sheet_element=self.getSheetElementByIndex(sheet_index)
        row=sheet_element.getElementsByType(TableRow)[coord.numberIndex()]
        return row.getElementsByType(TableCell)[coord.letterIndex()]
        
    ## Returns the cell value
    def getCellValue(self, sheet_index, coord):
        cell=self.getOdfPyCell(sheet_index, coord)
        return self._getCellValue_from_odfpy_cell(cell)
        
    ## Used to improve performance avoiding searching cells
    def _getCellValue_from_odfpy_cell(self, cell):
        r=None
        if cell.getAttribute('formula') is not None:
            r=Formula_from_OdfpyCell(cell)
        elif cell.getAttribute('valuetype')=='string':
            r=str(cell)
        elif cell.getAttribute('valuetype')=='float':
            r=Decimal(cell.getAttribute('value'))
        elif cell.getAttribute('valuetype')=='percentage':
            r=Percentage(Decimal(cell.getAttribute('value')), Decimal(1))
        elif cell.getAttribute('valuetype')=='currency':
            r=Currency(Decimal(cell.getAttribute('value')), cell.getAttribute('currency'))
        elif cell.getAttribute('valuetype')=='date':
            datevalue=cell.getAttribute('datevalue')
            if len(datevalue)<=10:
                r=datetime.strptime(datevalue, "%Y-%m-%d").date()
            else:
                r=datetime.strptime(datevalue, "%Y-%m-%dT%H:%M:%S")
        elif cell.getAttribute('valuetype')=='time':
            s=cell.getAttribute('timevalue')
            h=int(s.split("PT")[1].split("H")[0])
            m=int(s.split("H")[1].split("M")[0])
            s=int(s.split("M")[1].split("S")[0])
            r=time(h, m, s)
        else:
            return None
        return r

    ## Returns an odfcell object
    def getCell(self, sheet_index,  coord):
        cell=self.getOdfPyCell(sheet_index, coord)
        object=self._getCellValue_from_odfpy_cell(cell)
        #Get spanning
        spanning_columns=cell.getAttribute('numbercolumnsspanned')
        if spanning_columns==None:
            spanning_columns=1
        else:
            spanning_columns=int(spanning_columns)
        spanning_rows=cell.getAttribute('numberrowsspanned')
        if spanning_rows==None:
            spanning_rows=1
        else:
            spanning_rows=int(spanning_rows)
        
        #Get Stylename
        stylename=cell.getAttribute('stylename')


        #Odfcell
        r=OdfCell(coord, object, stylename)
        r.setSpanning(spanning_columns, spanning_rows)
        return r
        
    def setCell(self, sheet_index,  coord, odfcell):
        """
            odfcell is a officegenerator.OdfCell object
            Updates a cell
            insertBefore(newchild, refchild) – Inserts the node newchild before the existing child node refchild.
appendChild(newchild) – Adds the node newchild to the end of the list of children.
removeChild(oldchild) – Re

        ESTA FUNCION SE USA PARA SUSTITUIR EN UNA PLANTILLA
        NO SE PUEDEN AÑADIR MAS CELDAS O FILAS
        PARA ESO USAR ODS_Write DE MOMENTO
        """
        coord=Coord.assertCoord(coord)
        sheet_element=self.getSheetElementByIndex(sheet_index)
        row=sheet_element.getElementsByType(TableRow)[coord.numberIndex()]
        oldcell=row.getElementsByType(TableCell)[coord.letterIndex()]
        row.insertBefore(odfcell.generate(), oldcell)
        row.removeChild(oldcell)        



    def save(self, filename):
        if  filename==self.filename:
            print(_("You can't overwrite a readed ods"))
            return        
        self.doc.save( filename)


## Abstract class the defines opendocument metadata, images, languages
class ODF:
    def __init__(self, filename):
        self.filename=filename
        self.images={}
        
    ## Set metadata in odf document
    ## @param title String with the title of the document
    ## @param subject String with a brief of the document
    ## @param creator String with the author of the document
    ## @param description String with a larga description of the document
    ## @param keywords String with keywords separated by space
    ## @param creationdate Naive datetime with the creation date and time
    def setMetadata(self, title="",  subject="", creator="", description="", keywords="", creationdate=datetime.now()):
        for e in self.doc.meta.childNodes:
            self.doc.meta.removeChild(e)
        self.doc.meta.addElement(Description(text=description))
        self.doc.meta.addElement(Title(text=title))
        self.doc.meta.addElement(Subject(text=subject))
        self.doc.meta.addElement(Creator(text=creator))
        self.doc.meta.addElement(InitialCreator(text=creator))
        self.doc.meta.addElement(Keyword(text=keywords))
        d=Date()
        d.addText(creationdate.strftime("%Y-%m-%dT%H:%M:%S"))
        self.doc.meta.addElement(CreationDate(text=d))

    ## Adds an image to self.images dictionary. We add to a dictionary in order to reuse the same image in a directory
    ##
    ## One in the directory you can use the image with ODT.image()
    ## @param path String with the path to the image
    ## @param key String with the key of self.images dictionary. By default is None, in this case, the path string will be the key.
    def addImage(self, path, key=None):
        if key==None:
            key=path
        self.images[key]=self.doc.addPicture(path)

    def setLanguage(self, language, country):
        """Set the main language of the document"""
        self.language="es"
        self.country="ES"
        
    def showElement(self, e):
        print("ATTRIBUTE_NODE: {}".format(e.ATTRIBUTE_NODE))
        print("CDATA_SECTION_NODE: {}".format(e.CDATA_SECTION_NODE))
        print("TEXT_NODE: {}".format(e.TEXT_NODE))
        print("Atributes: {}".format(e.attributes))
        print("QNAME: {}".format(e.qname))
        print("tagName: {}".format(e.tagName))
        print("Allowed attributes: {}".format(e.allowed_attributes()))
        
        print(e)

## Class used to generate a ODT file with predefined formats
## @param filename String with the name of the filename to read, then will be saved with a different name
## @param template String with the name of the filename used as template
## @param language String with language. For example es
## @param country String with the country. For example ES
## @param predefinedstyles Boolean that sets if predefined styles are going to be loaded. True by default
class ODT(ODF):
    def __init__(self, filename, template=None, language="es", country="ES"):
        ODF.__init__(self, filename)
        self.setLanguage(language, country)
        ## After inserting an element it sets the new element as cursor to append 

        self.seqTables=0#Sequence of tables
        self.seqFrames=0#If a frame is repeated it doesn't show its
        self.template=template
        if self.template==None:
            self.doc=OpenDocumentText()
        else:
            self.doc= load(self.template)
            self.save()#After loading a template it's needed to save file to work with search functions
            remove(self.filename)#Delete temporal save
        self.cursor=None
        self.cursorParent=self.doc.text

    ## Creates a text header
    ## @param text String with the header string
    ## @Level Integer Level of the header
    def header(self, text, level, after=True):
        h=H(outlinelevel=level, stylename="Heading_20_{}".format(level), text=text)
        return self.insertInCursor(h, after)

    def pageBreak(self,  horizontal=False, after=True):    
        p=P(stylename="PageBreak")#Is an automatic style
        self.doc.text.addElement(p)
        if horizontal==True:
            p=P(stylename="PH")
        else:
            p=P(stylename="PV")
        return self.insertInCursor(p, after)


    ## @param href must bu added before with addImage
    ## @param width Int or float value
    ## @param height Int or float value to set the images height
    ## @return Frame element
    def image(self, href, width, height, name=None):
        self.seqFrames=self.seqFrames+1
        name=name if name!=None else "Frame.{}".format(self.seqFrames)

        f = Frame(name=name, anchortype="as-char", width="{}cm".format(width), height="{}cm".format(height))
        img = Image(href=self.images[href], type="simple", show="embed", actuate="onLoad")
        f.addElement(img)
        return f

    ## List_20_1 is a list style. Don't get wrong with List_20_1 paragraph style
    ## @param arr List of strings and other list. You must take care that level2 list are included in list-item of level 1.
    ## @param style is a list style not a paragraph style. In standard configured styles are "List_20_1" and "Numbering_20_123"
    ## @param boolean To insert before or after the current cursor.
    ## @code multilevel
    ##    doc.list(   [   ["1", ["1.1", "1.2"]], 
    ##                    ["2"], 
    ##                    ["3",  ["3.1", ["3.1.1", "3.1.2"]]]
    ##                ],  style="List_20_1")   
    ## @endcode
    ## @code unilevel
    ##    doc.list(   [   ["1",], 
    ##                        ["2",], 
    ##                        ["3",],
    ##                ],  style="List_20_1")   
    ## @endcode
    ## Se genera un nivel más de list no se porque pero queda bien.
    def list(self, arr, list_style="List_20_1", paragraph_style="Text_20_body", after=True):
        def get_items(list_o, list_style, paragraph_style):
            r=[]
            for o in list_o:
                it=ListItem()
                if o.__class__==str:
                    it.addElement(P(stylename=paragraph_style, text=o))
                else:
                    it.addElement(get_list(o, list_style, paragraph_style))
                r.append(it)
            return r
        def get_list(arr, list_style, paragraph_style):
            ls=List(stylename=list_style)
            for listitem in get_items(arr, list_style, paragraph_style):
                ls.addElement(listitem)
            return ls
        # #########################
        return self.insertInCursor(get_list(arr, list_style, paragraph_style), after)    

    ## Extracts odf document structure
    def odf_dump_nodes(self, start_node, level=0):
        if start_node.nodeType==3:
            # text node
            print ("  "*level, "NODE:", start_node.nodeType, ":(text):")
        else:
            # element node
            attrs= []
            for k in start_node.attributes.keys():
                attrs.append( str(k[1]) + ':' + str(start_node.attributes[k]  ))
            print("{} NODE: {}:{} ATTR: {}".format(" "*level,  str(start_node.nodeType), str(start_node.qname[1]), attrs))

            for n in start_node.childNodes:
                self.odf_dump_nodes(n, level+1)

    ## Adds a paragraph
    ## @param text to add
    ## @param style Style to use
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def simpleParagraph(self, text, style="Standard", after=True):
        p= P(stylename=style, text=text)
        return self.insertInCursor(p, after)

    ## Inserts after or before the Cursor, and sets the Cursor to the o element
    ## @param o Object to insert
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def insertInCursor(self, o, after):
        if after==None:# It doesn't insert, just return item
            return o

        if self.cursor==None:# First insert
            self.cursorParent.addElement(o)
        elif after==True:
            indexcursor=self.cursorParent.childNodes.index(self.cursor)
            if len(self.cursorParent.childNodes)==indexcursor+1:
                self.cursorParent.insertBefore(o, None)
            else:
                siguiente=self.cursorParent.childNodes[indexcursor+1]
                self.cursorParent.insertBefore(o, siguiente)
        else:#After = False
            self.cursorParent.insertBefore(o, self.cursor)
        self.__setCursor(o)
        return o

    ## Search for a tag type_ elementes  an returns element and its index. With it, inserts new with replaced text and removes old
    ##
    ## Cursor doesn't change because we replace Text objects in Element Text
    ## @param tag String to search
    ## @param replace String to replace. Can't be None
    ## @param Search in elements with tag P or H. To encapsulate type_ will be a string "P" or "H"
    def search_and_replace(self, tag, replace, type_="P"):
        e,  textindex=self.search(tag, type_) #Places cursor to element
        if e==None:
            return

        if replace==None:#Remove paragraph
            print("Replace parameter can't be None. Use '' instead or use search_and_replace_element")
            return

        #Replace text with the same style as found paragraph
        to_remove=e.childNodes[textindex]
        if replace==None:#Removes text
            e.removeChild(to_remove)
            if len(e.childNodes)==0:#Removes element
                self.cursorParent.removeChild(e)
        else: #Replace
            e.insertBefore( odf.element.Text(str(to_remove).replace(tag, replace)), to_remove)
            #print(to_remove.__class__, e.__class__, e.childNodes)
            debug("THIS CODE FAILS IN ODFPY-1.4.1 AND WORKS IN ODFPY-1.3.6")
            e.removeChild(to_remove     )
            

    ## Search for a tag in doc an replaces its elemente with the parameter element
    ##
    ## @param tag String to search
    ## @param replace ELement. OdfPy element
    def search_and_replace_element(self, tag, newelement, type_="P"):
        e,  textindex=self.search(tag, type_) #Places cursor to element
        if e==None:
            return

        if newelement==None:#Remove paragraph
            print("New element can't be None")
            return

        self.cursorParent.insertBefore(newelement,e)
        self.cursorParent.removeChild(e)
        self.__setCursor(newelement)


    ## Searchs for the item with a tag. Perhaps is its paren where I'll have to append. Only finds the first one
    ## Returns the element p and the position in its text children
    ## Using templates sometimes you cant search a tag. It's due to sometimes has <span> in the tag. Use odf2xml to detect them
    ## 20200119 Search_and_replace has problem with tables, we must try to not abuse of them
    def search(self, tag, type_="P"):
        if type_=="P":
             ty=P
        elif type_=="H":
             ty=H
        for e in self.doc.getElementsByType(ty):
            #print("Searching", tag, "found", str(e))
            if str(e).find(tag)!=-1:
                self.__setCursor(e)
                for index, child in enumerate(e.childNodes):
                    #print(index, child)
                    #print("Searching in for", tag,"-", child , "found", str(e))
                    if str(child).find(tag)!=-1:
                        #print("SEARCH RETURN",  e, index)
                        return e, index
        print ("Tag {} not found with type {}".format(tag,type_))
        return None, None


    ## Converts saved odt to pdf. It will have the same basename but with .pdf extension
    ## @param output_dir None or dirname. If set, it writes pdf in that directory with the same basename of the filename class attribute
    def convert_to_pdf(self, output_dir=None):
        if output_dir is None:
            s_outdir=""
        else:
            s_outdir="--outdir '{}'".format(output_dir)
        convert_command(self.filename, s_outdir, "pdf")

    def __setCursor(self, e):
        self.cursor=e
        self.cursorParent=e.parentNode
        if self.cursorParent==None:
            print("Parent of '{}' is None".format(self.cursor))

    ## Check if self.filename is different to self.template, create directory and saves the file
    def save(self):
        if  self.filename==self.template:
            print(_("You can't overwrite a readed odt"))
            return
        if path.dirname(self.filename)!="":
            makedirs(path.dirname(self.filename), exist_ok=True)
        self.doc.save( self.filename)

    ## Adds an empty paragraph
    ## @param style String with the style name
    ## @param number Integer with the number of times to repeat the empty paragraph
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def emptyParagraph(self, style="Standard", number=1, after=True):
        for i in range(number):
            self.simpleParagraph("",style, after)

    ## Creates the document title
    ## @param text String with the title
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def title(self, text, after=True):
        p=P(stylename="Title", text=text)
        return self.insertInCursor(p, after)


    ## Creates the document title
    ## @param text String with the title
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def subtitle(self, text, after=True):
        p=P(stylename="Subtitle", text=text)
        return self.insertInCursor(p, after)



    ## Creates a table adding it to self.doc
    ## @param hh List with all header strings
    ## @param data Multidimension List with all data objects. Can be str, Decimal, int, datetime, date, Currency, Percentage
    ## @param sizes Integer list with sizes in cm
    ## @param fontsize Integer in pt
    ## @param name str or None. Sets the object name. Appears in LibreOffice navigator. If none table will be named to "Table.Sequence"
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def table(self, hh, data, sizes, fontsize, name=None, after=True):
        def generate_table_styles():
            s=Style(name="Table.Size{}".format(sum(sizes)), family='table')
            s.addElement(TableProperties(width="{}cm".format(sum(sizes)), align="center", margintop="0.6cm", marginbottom="0.6cm"))
            self.doc.automaticstyles.addElement(s)

            #Column sizes
            for i, size in enumerate(sizes):
                sc= Style(name="Table.Column.Size{}".format(size), family="table-column")
                sc.addElement(TableColumnProperties(columnwidth="{}cm".format(size)))
                self.doc.automaticstyles.addElement(sc)

            #Cell header style
            sch=Style(name="Table.HeaderCell", family="table-cell")
            sch.addElement(TableCellProperties(backgroundcolor="#999999",  border="0.05pt solid #000000", padding="0.15cm"))
            self.doc.automaticstyles.addElement(sch)        

            #Cell normal
            sch=Style(name="Table.Cell", family="table-cell")
            sch.addElement(TableCellProperties(border="0.05pt solid #000000", padding="0.1cm"))
            self.doc.automaticstyles.addElement(sch)

            #TAble contents style
            s= Style(name="Table.Heading.Font{}".format(fontsize), family="paragraph" )
            s.addElement(TextProperties(attributes={'fontsize':"{}pt".format(fontsize),  'fontweight':"bold"}))
            s.addElement(ParagraphProperties(attributes={'textalign':'center', }))
            self.doc.styles.addElement(s)

            s = Style(name="Table.Contents.Font{}".format(fontsize), family="paragraph")
            s.addElement(TextProperties(attributes={'fontsize':"{}pt".format(fontsize), }))
            s.addElement(ParagraphProperties(attributes={'textalign':'justify', }))
            self.doc.styles.addElement(s)

            s = Style(name="Table.ContentsRight.Font{}".format(fontsize), family="paragraph")
            s.addElement(TextProperties(attributes={'fontsize':"{}pt".format(fontsize), }))
            s.addElement(ParagraphProperties(attributes={'textalign':'end', }))
            self.doc.styles.addElement(s)

            s = Style(name="Table.ContentsRight.FontRed{}".format(fontsize), family="paragraph")
            s.addElement(TextProperties(attributes={'fontsize':"{}pt".format(fontsize), 'color':'#ff0000' }))
            s.addElement(ParagraphProperties(attributes={'textalign':'end', }))
            self.doc.styles.addElement(s)

        ## Generate a TableCell guessing style, setting color if number is negative and setting alignment
        ## @param o Object can be str, Decimal, int, float, datetime, date, Currency, Percentage
        ## @param fontsize Integer with the size of the font
        ## @return TableCell
        def addTableCell(o, fontsize):
            tc = TableCell(stylename="Table.Cell")
            #Parses orientation
            p = P(stylename="Table.ContentsRight.Font{}".format(fontsize))
            s=Span(text=str(o))
            if o.__class__.__name__  == "datetime":
                p = P(stylename="Table.Contents.Font{}".format(fontsize))
                s=Span(text=dtnaive2string(o, "%Y-%m-%d %H:%M:%S"))
            if o.__class__.__name__ in ("str", "date" ):
                p = P(stylename="Table.Contents.Font{}".format(fontsize))
                s=Span(text=str(o))
            elif o.__class__.__name__ in ("Currency", "Percentage", "Money"):
                if o.isLTZero():
                    p = P(stylename="Table.ContentsRight.FontRed{}".format(fontsize))
                s=Span(text=o.string())
            elif o.__class__ in ("int", "Decimal",  "float"):
                if o<0:
                    p = P(stylename="Table.ContentsRight.FontRed{}".format(fontsize))
            p.addElement(s)
            tc.addElement(p)
            return tc

        ######################################
        self.seqTables=self.seqTables+1
        generate_table_styles()
        #Table columns
        name=name if name!=None else "Table.{}".format(self.seqTables)
        table = Table(name=name, stylename="Table.Size{}".format(sum(sizes)))
        for size in sizes:
            table.addElement(TableColumn(stylename="Table.Column.Size{}".format(size)))
        #Header rows
        headerrow=TableHeaderRows()
        tablerow=TableRow()
        headerrow.addElement(tablerow)
        if hh is not None:
            for i, head in enumerate(hh):
                p=P(stylename="Table.Heading.Font{}".format(fontsize), text=head)
                tablecell=TableCell(stylename="Table.HeaderCell")
                tablecell.addElement(p)
                tablerow.addElement(tablecell)
        table.addElement(headerrow)

        #Data rows
        for row in data:
            tr = TableRow()
            table.addElement(tr)
            for i, o in enumerate(row):
                tr.addElement(addTableCell(o, fontsize))
        return self.insertInCursor(table, after)

    ## Adds a paragraph with illustration style, with a list with image keys. All images will have the same width and height
    ## @param image_key_list List with image keys to add to the paragraph
    ## @param width Integer or float with number of width centimeters of the image in the document
    ## @param height Integer or float with number of height centimeters of the image in the document
    ## @param name String With the root name of the images in the paragraph. The root will be appended with a sequential number
    ## @param after True: insert after self.cursor element. False: insert before self.cursor element. None: Just return element
    def illustration(self, image_key_list, width, height, name, after=True):
        p = P(stylename="Illustration")
        for i, image_key in enumerate(image_key_list):
            if name==None:
                self.seqFrames=self.seqFrames+1
                n="Frame.{}".format(self.seqFrames)
            else:
                n="{}.{}".format(name, i)
            p.addElement(self.image(image_key, width, height, n))
        self.insertInCursor(p, after=True)

## Class with starndard.odt template
class ODT_Standard(ODT):
    def __init__(self, filename, language="es", country="es"):
        template=resource_filename("officegenerator","templates/odt/standard.odt")
        ODT.__init__(self, filename, template, language, country)

    ## PH and PV styles are deffined in standard.odtg
    def pageBreak(self,  horizontal=False, after=True):    
        if horizontal==True:
            p=P(stylename="PH")
        else:
            p=P(stylename="PV")
        return self.insertInCursor(p, after)

## Class with starndard.odt template
class ODT_Manual_Styles(ODT):
    def __init__(self, filename, language="es", country="es"):
        ODT.__init__(self, filename, None, language, country)
        self.load_predefined_styles()

    ## Creates a text header
    ## @param text String with the header string
    ## @Level Integer Level of the header
    def header(self, text, level, after=True):
        h=H(outlinelevel=level, stylename="Heading{}".format(level), text=text)
        return self.insertInCursor(h, after)
    ## Loads predefined styles. If you want to change them you need to make a class with ODT as its parent and override this method or to load a template
    def load_predefined_styles(self):
        def stylePage():
            pagelayout=PageLayout(name="PageLayout")
            plp=PageLayoutProperties(pagewidth="21cm",  pageheight="29.7cm",  margintop="2cm",  marginright="2cm",  marginleft="2cm",  marginbottom="2cm")
            fs=FooterStyle()
            hfp=HeaderFooterProperties(margintop="0.5cm")
            fs.addElement(hfp)
            pagelayout.addElement(plp)
            pagelayout.addElement(fs)
            self.doc.automaticstyles.addElement(pagelayout)

        def styleParagraphs():
            #Pagebreak styles horizontal y vertical        
            s = Style(name="PH", family="paragraph",  parentstylename="Standard", masterpagename="Landscape")
            s.addElement(ParagraphProperties(pagenumber="auto"))
            self.doc.styles.addElement(s)
            s = Style(name="PV", family="paragraph",  parentstylename="Standard", masterpagename="Standard")
            s.addElement(ParagraphProperties(pagenumber="auto"))
            self.doc.styles.addElement(s)

            standard= Style(name="Standard", family="paragraph",  autoupdate="true")
            standard.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"justify", "marginbottom":"0.2cm", "textindent":"1.5cm"}))
            standard.addElement(TextProperties(attributes={"fontsize": "12pt", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(standard)   

            ImageCenter= Style(name="Illustration", family="paragraph",  autoupdate="true")
            ImageCenter.addElement(ParagraphProperties(attributes={"margintop":"1cm", "textalign":"center", "marginbottom":"1cm"}))
            ImageCenter.addElement(TextProperties(fontsize="12pt", fontstyle="italic", country=self.country, language=self.language))
            self.doc.styles.addElement(ImageCenter)

            standardCenter= Style(name="StandardCenter", family="paragraph",  autoupdate="true")
            standardCenter.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"center", "marginbottom":"0.2cm", "textindent":"0cm"}))
            standardCenter.addElement(TextProperties(attributes={"fontsize": "12pt", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(standardCenter)

            standardRight= Style(name="StandardRight", family="paragraph",  autoupdate="true")
            standardRight.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"right", "marginbottom":"0.2cm", "textindent":"0cm"}))
            standardRight.addElement(TextProperties(attributes={"fontsize": "12pt", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(standardRight)



            letra12= Style(name="Bold12Underline", family="paragraph",  autoupdate="true")
            letra12.addElement(ParagraphProperties(attributes={"margintop":"0.4cm", "textalign":"justify", "marginbottom":"0.4cm", "textindent":"1.5cm"}))
            letra12.addElement(TextProperties(attributes={"fontsize": "12pt", "fontweight": "bold", "country": self.country, "language": self.language, 'textunderlinecolor':"font-color", 'textunderlinewidth':"auto", 'textunderlinestyle':"solid"}))
            self.doc.styles.addElement(letra12)


            letra18= Style(name="Bold18Center", family="paragraph",  autoupdate="true")
            letra18.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"center", "marginbottom":"0.2cm", "textindent":"0cm"}))
            letra18.addElement(TextProperties(attributes={"fontsize": "18pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(letra18)

            letra16= Style(name="Bold16Center", family="paragraph",  autoupdate="true")
            letra16.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"center", "marginbottom":"0.2cm", "textindent":"0cm"}))
            letra16.addElement(TextProperties(attributes={"fontsize": "16pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(letra16)

            letra14= Style(name="Bold14Center", family="paragraph",  autoupdate="true")
            letra14.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"center", "marginbottom":"0.2cm", "textindent":"0cm"}))
            letra14.addElement(TextProperties(attributes={"fontsize": "14pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(letra14)

            letra12= Style(name="Bold12Center", family="paragraph",  autoupdate="true")
            letra12.addElement(ParagraphProperties(attributes={"margintop":"0.2cm", "textalign":"center", "marginbottom":"0.2cm", "textindent":"0cm"}))
            letra12.addElement(TextProperties(attributes={"fontsize": "12pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(letra12)

        def styleHeaders():
            s = Style(name="Title", family="paragraph",  autoupdate="true", defaultoutlinelevel="0")
            s.addElement(ParagraphProperties(attributes={"margintop":"0.7cm", "textalign":"center", "marginbottom":"0.7cm"}))
            s.addElement(TextProperties(attributes={"fontsize": "18pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(s)

            s = Style(name="Subtitle", family="paragraph",  autoupdate="true", defaultoutlinelevel="0")
            s.addElement(ParagraphProperties(attributes={"margintop":"0cm", "textalign":"center", "marginbottom":"0.7cm"}))
            s.addElement(TextProperties(attributes={"fontsize": "15pt", "fontstyle": "italic", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(s)

            h1style = Style(name="Heading1", family="paragraph",  autoupdate="true", defaultoutlinelevel="1")
            h1style.addElement(ParagraphProperties(attributes={"margintop":"0.6cm", "textalign":"justify", "marginbottom":"0.3cm"}))
            h1style.addElement(TextProperties(attributes={"fontsize": "15pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(h1style)

            h2style = Style(name="Heading2", family="paragraph",  autoupdate="true", defaultoutlinelevel="2")
            h2style.addElement(ParagraphProperties(attributes={"margintop":"0.5cm", "textalign":"justify", "marginbottom":"0.25cm"}))
            h2style.addElement(TextProperties(attributes={"fontsize": "14pt", "fontweight": "bold", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(h2style)

            out=OutlineStyle(name="Outline")
            outl=OutlineLevelStyle(level=1, numformat="1", numsuffix="  ")
            out.addElement(outl)
            outl=OutlineLevelStyle(level=2, displaylevels="2", numformat="1", numsuffix="  ")
            out.addElement(outl)
            self.doc.styles.addElement(out)

        def styleList():
            liststandard= Style(name="ListStandard", family="paragraph",  autoupdate="true")
            liststandard.addElement(ParagraphProperties(attributes={"margintop":"0.1cm", "textalign":"justify", "marginbottom":"0.1cm", "textindent":"0cm"}))
            liststandard.addElement(TextProperties(attributes={"fontsize": "12pt", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(liststandard)
            
            # For Bulleted list
            bulletedliststyle = ListStyle(name="BulletList1")
            bulletlistproperty = ListLevelStyleBullet(level="1", bulletchar=u"•")
            bulletlistproperty.addElement(ListLevelProperties( minlabelwidth="1cm"))
            bulletedliststyle.addElement(bulletlistproperty)
            self.doc.styles.addElement(bulletedliststyle)
            # For Bulleted list
            bulletedliststyle = ListStyle(name="BulletList2")
            bulletlistproperty = ListLevelStyleBullet(level="2", bulletchar=u"#")
            bulletlistproperty.addElement(ListLevelProperties( minlabelwidth="1cm"))
            bulletedliststyle.addElement(bulletlistproperty)
            self.doc.styles.addElement(bulletedliststyle)
            # For Bulleted list
            bulletedliststyle = ListStyle(name="BulletList2")
            bulletlistproperty = ListLevelStyleBullet(level="2", bulletchar=u"·")
            bulletlistproperty.addElement(ListLevelProperties( minlabelwidth="1cm"))
            bulletedliststyle.addElement(bulletlistproperty)
            self.doc.styles.addElement(bulletedliststyle)

            # For numbered list
            numberedliststyle = ListStyle(name="NumberedList")
            numberedlistproperty = ListLevelStyleNumber(level="1", numsuffix=".", startvalue=1)
            numberedlistproperty.addElement(ListLevelProperties(minlabelwidth="1cm"))
            numberedliststyle.addElement(numberedlistproperty)
            self.doc.styles.addElement(numberedliststyle)
        
        def styleFooter():
            s= Style(name="Footer", family="paragraph",  autoupdate="true")
            s.addElement(ParagraphProperties(attributes={"margintop":"0cm", "textalign":"center", "marginbottom":"0cm", "textindent":"0cm"}))
            s.addElement(TextProperties(attributes={"fontsize": "9pt", "country": self.country, "language": self.language}))
            self.doc.styles.addElement(s)

        def styleMasterPage():
            foot=MasterPage(name="Standard", pagelayoutname="PageLayout")
            footer=Footer()
            p1=P(stylename="Footer",  text="Página ")
            number=PageNumber(selectpage="current", numformat="1")
            p2=Span(stylename="Footer",  text=" de  ")
            count=PageCount(selectpage="current", numformat="1")
            p1.addElement(number)
            p1.addElement(p2)
            p1.addElement(count)      
            footer.addElement(p1)   
            foot.addElement(footer)
            self.doc.masterstyles.addElement(foot)
        #######################################                
        stylePage()
        styleParagraphs()
        styleFooter()
        styleMasterPage()
        styleHeaders()
        styleList()

    ## This function always use manual Bulletlist and liststandard styles by coherence with manual styles
    def list(self, arr, list_style=None, paragraph_style=None,  after=True):
        return ODT.list(self, arr, "BulletList", "ListStandard", after)

    def pageBreak(self,  horizontal=False, after=True):    
        p=P(stylename="PageBreak")#Is an automatic style
        self.doc.text.addElement(p)
        if horizontal==True:
            p=P(stylename="PH")
        else:
            p=P(stylename="PV")
        return self.insertInCursor(p, after)

## Manage Odf Cells
class OdfCell:
    def __init__(self, coord,  object, style):
        self.coord=Coord.assertCoord(coord)
        self.object=object
        self.style=style

        self.spannedColumns=1
        self.spannedRows=1
        self.comment=None

    def __repr__(self):
        return "OdfCell <{}{}>".format(self.coord.letter, self.coord.number)

    def generate(self):
        if self.object==None:
            self.object=" - "
        if self.object.__class__.__name__ in ["Currency", "Money"]:
            odfpycell = TableCell(valuetype="currency", currency=self.object.currency, value=self.object.amount, stylename=self.style)
        elif self.object.__class__.__name__=="Percentage":
            odfpycell = TableCell(valuetype="percentage", value=self.object.value, stylename=self.style)
        elif self.object.__class__.__name__=="datetime":
            odfpycell = TableCell(valuetype="date", datevalue=self.object.strftime("%Y-%m-%dT%H:%M:%S"), stylename=self.style)
        elif self.object.__class__.__name__=="time":
            odfpycell = TableCell(valuetype="time", timevalue=self.object.strftime("PT%HH%MM%SS"), stylename=self.style)
        elif self.object.__class__.__name__=="date":
            odfpycell = TableCell(valuetype="date", datevalue=str(self.object), stylename=self.style)
        elif self.object.__class__.__name__ in ("Decimal", "float", "int"):
            odfpycell= TableCell(valuetype="float", value=self.object,  stylename=self.style)
        elif self.object.__class__.__name__ =="bool":
            odfpycell= TableCell(valuetype="boolean", booleanvalue=self.object, stylename=self.style)
        elif self.object.__class__.__name__ =="Formula":
            odfpycell = TableCell(formula="of:"+self.object.string_formula,  stylename=self.style)
        else:#strings
            odfpycell = TableCell(valuetype="string", value=self.object,  stylename=self.style)
            odfpycell.addElement(P(text = self.object))

        if self.spannedRows!=1 or self.spannedColumns!=1:
            odfpycell.setAttribute("numberrowsspanned", str(self.spannedRows))
            odfpycell.setAttribute("numbercolumnsspanned", str(self.spannedColumns))
        if self.comment!=None:
            a=Annotation(textstylename="Right")
            d=Date()
            d.addText(datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
            a.addElement(d)
            a.addElement(P(stylename="Right", text=self.comment))
            odfpycell.addElement(a)
        return odfpycell

    ## Manage cell spannning 
    ##
    ## Siempre es de izquierda a derecha y de arriba a abajo. Si es 1 no hay spanning
    def setSpanning(self, columns, rows):
        self.spannedColumns=columns
        self.spannedRows=rows

    def setComment(self, comment):
        self.comment=comment


## Class to create a sheet in a ods document
## By default cursor position and split position is set to "A1" cell
class OdfSheet:
    def __init__(self, doc,  title):
        self.doc=doc
        self.title=title
        self.widths=None
        self.arr=[]
        self.freezeAndSelect("A1", "A1", "A1")#Default values

    ## Freeze panels in a sheet and sets the selected cell
    ##        split/freeze vertical (0|1|2) - 1 = split ; 2 = freeze
    ##          ##split/freeze horizontal (0|1|2) - 1 = split ; 2 = freeze
    ##      vertical position = in cell if fixed, in screen unit if frozen
    ##      horizontal position = in cell if fixed, in screen unit if frozen
    ##      active zone in the splitted|frozen sheet (0..3 from let to right, top
    ##  to bottom)
    ##  #   COMPROBADO CON ODF2XML
    ##  B1: 
    ##                <config:config-item config:name="HorizontalSplitMode" config:type="short">2</config:config-item>
    ##                <config:config-item config:name="VerticalSplitMode" config:type="short">0</config:config-item>
    ##                <config:config-item config:name="HorizontalSplitPosition" config:type="int">1</config:config-item>
    ##                <config:config-item config:name="VerticalSplitPosition" config:type="int">0</config:config-item>
    ##                <config:config-item config:name="ActiveSplitRange" config:type="short">3</config:config-item>
    ##                <config:config-item config:name="PositionLeft" config:type="int">0</config:config-item>
    ##                <config:config-item config:name="PositionRight" config:type="int">1</config:config-item>
    ##                <config:config-item config:name="PositionTop" config:type="int">0</config:config-item>
    ##                <config:config-item config:name="PositionBottom" config:type="int">0</config:config-item>
    ## @param freeze_coord, Cell where panels are frrozen. Can be a string or a Coord object.
    ## @param selected_coord. Cell selected opening sheet. Can be a string or a Coord object.
    ## @param topLeftCell, topleftcell to show in sheet after opening. Can be a string or a Coord object.
    def freezeAndSelect(self, freeze_coord, selected_coord=None, topleftcell_coord=None):
        if selected_coord is None:
            selected_coord=Coord(self.lastColumn()+self.lastRow())
        
        if topleftcell_coord is None:
            topleftcell_coord=topLeftCellNone(freeze_coord, selected_coord)
        # Creates Coord objects
        freeze_coord=Coord.assertCoord(freeze_coord)
        selected_coord=Coord.assertCoord(selected_coord)
        topleftcell_coord=Coord.assertCoord(topleftcell_coord)

        #Sets cursor position
        self.cursorPositionX=selected_coord.letterIndex()
        self.cursorPositionY=selected_coord.numberIndex()
    
        #Sets freeze position and modes
        self.horizontalSplitPosition=str(freeze_coord.letterIndex())
        self.verticalSplitPosition=str(freeze_coord.numberIndex())
        self.horizontalSplitMode="0" if self.horizontalSplitPosition=="0" else "2"
        self.verticalSplitMode="0" if self.verticalSplitPosition=="0" else "2"
        
        #Sets active split range and top left cell position
        if self.horizontalSplitPosition!="0" and self.verticalSplitPosition=="0":#C1 WORKS
            self.activeSplitRange="3"
            self.positionTop="0"
            self.positionBottom=str(topleftcell_coord.numberIndex())
            self.positionLeft="0"
            self.positionRight=str(topleftcell_coord.letterIndex())
        elif self.horizontalSplitPosition=="0" and self.verticalSplitPosition!="0":#A3 WORKS
            self.activeSplitRange="2"
            self.positionTop="0"
            self.positionBottom=str(topleftcell_coord.numberIndex())
            self.positionLeft=str(topleftcell_coord.letterIndex())
            self.positionRight="0"
        elif self.horizontalSplitPosition!="0" and self.verticalSplitPosition!="0": # C3  WORKS
            self.activeSplitRange="3"
            self.positionTop="0"
            self.positionBottom=str(topleftcell_coord.numberIndex())
            self.positionLeft="0"
            self.positionRight=str(topleftcell_coord.letterIndex())
        else:#A1 WORKS
            self.activeSplitRange="2"
            self.positionTop="0"
            self.positionBottom=str(topleftcell_coord.numberIndex())
            self.positionLeft=str(topleftcell_coord.letterIndex())
            self.positionRight=str(topleftcell_coord.letterIndex())

    ## Sets a comment in the givven cell
    ## @param coord can be Coord o Coord.string()
    ## @param comment String to insert as a comment in the cell
    def setComment(self, coord, comment):
        coord=Coord.assertCoord(coord)
        c=self.getCell(coord)
        c.setComment(comment)


    def setColumnsWidth(self, widths, unit="pt"):
        """
            widths is an int array
            id es el id del sheet de python
        """
        for w in widths:
            s=Style(name="{}_{}".format(id(self), w), family="table-column")
            s.addElement(TableColumnProperties(columnwidth="{}{}".format(w, unit)))
            self.doc.automaticstyles.addElement(s)   
        self.widths=widths
        
    ## Returns the last letter used in the sheet . Returns a string with the letter name of the column
    def lastColumn(self):
        return number2column(self.columns())

    ## Returns the last  row name used
    ## @return string row name
    def lastRow(self):
        return  number2row(self.rows())

    def addCell(self, cell): 
        self.arr.append(cell)
        
    ## Returns the cell in the Coord
    ## @return OdfCell
    def getCell(self, coord):
        coord=Coord.assertCoord(coord)
        for c in self.arr:
            if c.coord==coord:
                return c
        return None
            
    def getCellValue(self, coord):
        cell=self.getCell(coord)
        if cell==None:
            return None
        else:
            return self.getCell(coord).object
    
    ## Returns true if value is a string beginning with = or +
    ## @param value must be a string
    ## @return boolean
    def isFormula(self, value):
        if len(value)>0 and value[0] in ["=", "+"]:
            return True
        return False

    ## Adds a cell to the sheet using its coord, an object and a color or a style
    ## @param coord Coord where the cell is going to be created
    ## @param result Object to add to the Cell. Can be int, str, float, datetime, date, Currency, Percentage, Decimal, None (will be converted to " - ")
    ## @param color_or_style String with a color: Normal, White, Yellow, Orange, Blue, Red, GrayLight, GrayDark. Or a style WhiteInteger, YellowLeft, OrangeCenter, OrangeEUR, RedPercentage...
    def add(self, coord, result, color_or_style="Normal"):
        coord=Coord.assertCoord(coord)
        if isFormula(result):
            result=Formula(result)
        

        if result.__class__ in (list,):#Una lista
            for i,row in enumerate(result):
                if row.__class__ in (list, ):#Una lista de varias columnas
                    for j, column in enumerate(row):
                        style=guess_ods_style(color_or_style, result[i][j])
                        self.addCell(OdfCell(Coord(coord.string()).addColumn(j).addRow(i), result[i][j], style))
                else: #Any value not list if row.__class__ in (int, str, float, datetime,  date, Currency, Percentage,  Decimal, bool):#Una lista de una columna
                    style=guess_ods_style(color_or_style, result[i])
                    self.addCell(OdfCell(Coord(coord.string()).addRow(i), result[i], style))
        else: #Any value not list#result.__class__ in (str, int, float, datetime, date,  Currency, Percentage, Decimal,bool):#Un solo valor
            style=guess_ods_style(color_or_style, result)  
            self.addCell(OdfCell(coord, result, style))

    ## Adds a cell to self.arr with merge, content and style information
    ## @param range Range
    def addMerged(self, range, result, style):
        range=Range.assertRange(range)
        self.add(range.start, result, style)      
        c=self.getCell(range.start)
        c.setSpanning(range.numColumns(), range.numRows())
            
    ## @param cood Coord from we are going to add totals
    ## @param list_of_totals List with strings or keys. Example: ["Total", "#SUM", "#AVG"]...
    ## @param list_of_styles List with string styles or None. If none tries to guest from top column object. List example: ["GrayLightPercentage", "GrayLightInteger"]
    ## @param string with the row where th3e total begins
    ## @param string with the rew where the formula ends. If None it's a coord.row -1
    def addTotalsHorizontal(self, coord, list_of_totals, list_of_styles=None, row_from="2", row_to=None):
        coord=Coord.assertCoord(coord)
        for letter, total in enumerate(list_of_totals):
            coord_total=coord.addColumnCopy(letter)
            coord_total_from=Coord(coord_total.letter+row_from)
            if row_to is None:
                coord_total_to=coord_total.addRowCopy(-1)# row above
            else:
                coord_total=Coord(coord_total.letter+row_to)

            if list_of_styles is None:
                style=guess_ods_style("GrayLight", self.getCellValue(coord_total_from))
            else:
                style=list_of_styles[letter]

            self.add(coord_total, generate_formula_total_string(total, coord_total_from, coord_total_to),  style)
            
    ## @param cood Coord from we are going to add totals
    ## @param list_of_totals List with strings or keys. Example: ["Total", "#SUM", "#AVG"]...
    ## @param list_of_styles List with string styles or None. If none tries to guest from top column object. List example: ["GrayLightPercentage", "GrayLightInteger"]
    ## @param string with the row where th3e total begins
    ## @param string with the rew where the formula ends. If None it's a coord.row -1
    def addTotalsVertical(self, coord, list_of_totals, list_of_styles=None, column_from="B", column_to=None):
        coord=Coord.assertCoord(coord)
        for i, total in enumerate(list_of_totals):
            coord_total=coord.addRowCopy(i)
            coord_total_from=Coord(column_from+coord_total.number)
            if column_to is None:
                coord_total_to=coord_total.addColumnCopy(-1)# row above
            else:
                coord_total=Coord(column_to+coord_total.number)

            if list_of_styles is None:
                style=guess_ods_style("GrayLight", self.getCellValue(coord_total_from))
            else:
                style=list_of_styles[i]

            self.add(coord_total, generate_formula_total_string(total, coord_total_from, coord_total_to),  style)

    ## Generates the sheet in self.doc Opendocument varianble
    def generate(self, ods):
        # Start the table
        columns=self.columns()
        rows=self.rows()
        grid=[[None for x in range(columns)] for y in range(rows)]
        for cell in self.arr:
            grid[cell.coord.numberIndex()][cell.coord.letterIndex()]=cell

        table = Table(name=self.title)
        for c in range(columns):#Create columns
            try:
                tc=TableColumn(stylename="{}_{}".format(id(self), self.widths[c]))
            except:
                tc=TableColumn()
            table.addElement(tc)
        for j in range(rows):#Crreate rows
            tr = TableRow()
            table.addElement(tr)
            for i in range(columns):#Create cells
                cell=grid[j][i]
                if cell is not None:
                    tr.addElement(cell.generate())
                else:
                    tr.addElement(TableCell())
        ods.doc.spreadsheet.addElement(table)

    ## Gets the number of columns that are used in the sheet
    ## @return int
    def columns(self):
        r=0
        for cell in self.arr:
            column=cell.coord.letterPosition()
            if column>r:
                r=column
        return r

    ## Return the number of rows that are used in the cell.
    def rows(self):
        r=0
        for cell in self.arr:
            column=cell.coord.numberPosition()
            if column>r:
                r=column
        return r


## Abstract class with common functions to generate Libreoffice Calc ODS files
class ODS(ODF):
    def __init__(self, filename):
        ODF.__init__(self, filename)
        self.doc=OpenDocumentSpreadsheet()
        self.sheets=[]
        self.activeSheet=None

    def createSheet(self, title):
        s=OdfSheet(self.doc, title)
        self.sheets.append(s)
        return s

    def getActiveSheet(self):
        if self.activeSheet==None:
            return self.sheets[0].title
        return self.activeSheet

    ## Save ODS file
    ## @param filename str or None. If filename is given, file is saved with a different name to the constructor one
    def save(self, filename=None):
        #config settings information
        a=ConfigItemSet(name="ooo:view-settings")
        aa=ConfigItem(type="int", name="VisibleAreaTop")
        aa.addText("0")
        a.addElement(aa)
        aa=ConfigItem(type="int", name="VisibleAreaLeft")
        aa.addText("0")
        a.addElement(aa)
        b=ConfigItemMapIndexed(name="Views")
        c=ConfigItemMapEntry()
        d=ConfigItem(name="ViewId", type="string")
        d.addText("view1")#value="view1"
        e=ConfigItemMapNamed(name="Tables")
        for sheet in self.sheets:
            f=ConfigItemMapEntry(name=sheet.title)
            g=ConfigItem(type="int", name="CursorPositionX")
            g.addText(sheet.cursorPositionX)
            f.addElement(g)
            g=ConfigItem(type="int", name="CursorPositionY")
            g.addText(sheet.cursorPositionY)
            f.addElement(g)
            g=ConfigItem(type="int", name="HorizontalSplitPosition")
            g.addText(sheet.horizontalSplitPosition)
            f.addElement(g)
            g=ConfigItem(type="int", name="VerticalSplitPosition")
            g.addText(sheet.verticalSplitPosition)
            f.addElement(g)
            g=ConfigItem(type="short", name="HorizontalSplitMode")
            g.addText(sheet.horizontalSplitMode)
            f.addElement(g)
            g=ConfigItem(type="short", name="VerticalSplitMode")
            g.addText(sheet.verticalSplitMode)
            f.addElement(g)
            g=ConfigItem(type="short", name="ActiveSplitRange")
            g.addText(sheet.activeSplitRange)
            f.addElement(g)
            g=ConfigItem(type="int", name="PositionLeft")
            g.addText(sheet.positionLeft)
            f.addElement(g)
            g=ConfigItem(type="int", name="PositionRight")
            g.addText(sheet.positionRight)
            f.addElement(g)
            g=ConfigItem(type="int", name="PositionTop")
            g.addText(sheet.positionTop)
            f.addElement(g)
            g=ConfigItem(type="int", name="PositionBottom")
            g.addText(sheet.positionBottom)
            f.addElement(g)
            e.addElement(f)
            
        a.addElement(b)
        b.addElement(c)
        c.addElement(d)
        c.addElement(e)

        h=ConfigItem(type="string", name="ActiveTable")
        h.addText(self.getActiveSheet())
        c.addElement(h)
        self.doc.settings.addElement(a)
        
        for sheet in self.sheets:
            sheet.generate(self)
        
        if  filename==None:
            filename=self.filename

        if path.dirname(filename)!="":
            makedirs(path.dirname(filename), exist_ok=True)
        self.doc.save(filename)

    def setActiveSheet(self, value):
        """value is OdfSheet"""
        self.activeSheet=value.title

class ODSStyleCurrency:
    def __init__(self, name, symbol):
        self.name=name
        self.symbol=symbol
        
    ## Generate currency styles
    def generate_ods_styles(self, doc):
        # Create the styles for $AUD format currency values
        ns1 = CurrencyStyle(name=self.name + "Black", volatile="true")
        ns1.addElement(Number(decimalplaces="2", minintegerdigits="1", grouping="true"))
        ns1.addElement(CurrencySymbol(language="es", country="ES", text=" "+ self.symbol))
        doc.styles.addElement(ns1)

        # Create the main style.
        ns2 = CurrencyStyle(name=self.name)
        ns2.addElement(TextProperties(color="#ff0000"))
        ns2.addElement(Text(text="-"))
        ns2.addElement(Number(decimalplaces="2", minintegerdigits="1", grouping="true"))
        ns2.addElement(CurrencySymbol(language="es", country="ES", text=" "+ self.symbol))
        ns2.addElement(Map(condition="value()>=0", applystylename=self.name + "Black"))
        doc.styles.addElement(ns2)
        

## Class to manage color in Libreoffice Calc ODS. It generates all needed ODS styles needed for that color
class ODSStyleColor:
    def __init__(self, name, rgb, bold):
        self.name=name
        self.rgb=rgb
        self.bold=bold

    def generate_ods_styles(self, doc, currencymanager):
        hs=Style(name=self.name + "Center", family="table-cell")
        hs.addElement(TableCellProperties(backgroundcolor=self.rgb, border="0.06pt solid #000000", verticalalign="middle", textalignsource="fix"))
        if self.bold==True:
            hs.addElement(TextProperties( fontweight="bold"))
        hs.addElement(ParagraphProperties(textalign="center"))
        doc.styles.addElement(hs)

        hs=Style(name=self.name + "Left", family="table-cell")
        hs.addElement(TableCellProperties(backgroundcolor=self.rgb, border="0.06pt solid #000000", verticalalign="middle", textalignsource="fix"))
        if self.bold==True:
            hs.addElement(TextProperties( fontweight="bold"))
        hs.addElement(ParagraphProperties(textalign="left"))
        doc.styles.addElement(hs)

        hs=Style(name=self.name + "Right", family="table-cell")
        hs.addElement(TableCellProperties(backgroundcolor=self.rgb, border="0.06pt solid #000000", verticalalign="middle", textalignsource="fix"))
        if self.bold==True:
            hs.addElement(TextProperties( fontweight="bold"))
        hs.addElement(ParagraphProperties(textalign="end"))
        doc.styles.addElement(hs)

        for currency in currencymanager.arr:
            moneycontents = Style(name=self.name+currency.name, family="table-cell",  datastylename=currency.name ,parentstylename=self.name+"Right")
            doc.styles.addElement(moneycontents)

        pourcent = Style(name=self.name+'Percentage', family='table-cell', datastylename='Percentage',parentstylename=self.name+"Right")
        doc.styles.addElement(pourcent)

        dt = Style(name=self.name+"Datetime", datastylename="Datetime",parentstylename=self.name+"Left", family="table-cell")
        doc.styles.addElement(dt)

        dat = Style(name=self.name+"Date", datastylename="Date",parentstylename=self.name+"Left", family="table-cell")
        doc.styles.addElement(dat)
        
        time = Style(name=self.name+"Time", datastylename="Time",parentstylename=self.name+"Left", family="table-cell")
        doc.styles.addElement(time)

        integer = Style(name=self.name+"Integer", family="table-cell",  datastylename="Integer",parentstylename=self.name+"Right")
        doc.styles.addElement(integer)
        
        boolean = Style(name=self.name+"Boolean", family="table-cell",  datastylename="Boolean",parentstylename=self.name+"Right")
        doc.styles.addElement(boolean)

        decimal2= Style(name=self.name+"Decimal2", family="table-cell",  datastylename="Decimal2",parentstylename=self.name+"Right")
        doc.styles.addElement(decimal2)

        decimal6= Style(name=self.name+"Decimal6", family="table-cell",  datastylename="Decimal6",parentstylename=self.name+"Right")
        doc.styles.addElement(decimal6)


## Class to Mange ODSStyleColors in libodfgenerator
class ODSStyleCurrencyManager:
    def __init__(self):
        self.arr=[]

    def append(self, o):
        self.arr.append(o)

    def generate_ods_styles(self, doc):
        for o in self.arr:
            o.generate_ods_styles(doc)

## Class to Mange ODSStyleColors in libodfgenerator
class ODSStyleColorManager:
    def __init__(self):
        self.arr=[]

    def append(self, o):
        self.arr.append(o)

    ## Generate common styles used in all color styles
    def __generate_ods_common_styles(self, doc):
        #Percentage
        nonze = PercentageStyle(name='PercentageBlack')
        nonze.addElement(TextProperties(color="#000000"))
        nonze.addElement(Number(decimalplaces='2', minintegerdigits='1'))
        nonze.addElement(Text(text=' %'))
        doc.styles.addElement(nonze)
        
        nonze2 = PercentageStyle(name='Percentage')
        nonze2.addElement(TextProperties(color="#ff0000"))
        nonze2.addElement(Text(text="-"))
        nonze2.addElement(Number(decimalplaces='2', minintegerdigits='1'))
        nonze2.addElement(Text(text=' %'))
        nonze2.addElement(Map(condition="value()>=0", applystylename="PercentageBlack"))
        doc.styles.addElement(nonze2)
        
        # Datetimes
        date_style = DateStyle(name="Datetime") #, language="lv", country="LV")
        date_style.addElement(Year(style="long"))
        date_style.addElement(Text(text="-"))
        date_style.addElement(Month(style="long"))
        date_style.addElement(Text(text="-"))
        date_style.addElement(Day(style="long"))
        date_style.addElement(Text(text=" "))
        date_style.addElement(Hours(style="long"))
        date_style.addElement(Text(text=":"))
        date_style.addElement(Minutes(style="long"))
        date_style.addElement(Text(text=":"))
        date_style.addElement(Seconds(style="long"))
        doc.styles.addElement(date_style)#NO SERIA ESTE UN AUTOMATICO????        
        
        # Time
        time_style = TimeStyle(name="Time") #, language="lv", country="LV")
        time_style.addElement(Hours(style="long"))
        time_style.addElement(Text(text=":"))
        time_style.addElement(Minutes(style="long"))
        time_style.addElement(Text(text=":"))
        time_style.addElement(Seconds(style="long"))
        doc.styles.addElement(time_style)#NO SERIA ESTE UN AUTOMATICO????

        #Date
        date_style = DateStyle(name="Date") #, language="lv", country="LV")
        date_style.addElement(Year(style="long"))
        date_style.addElement(Text(text="-"))
        date_style.addElement(Month(style="long"))
        date_style.addElement(Text(text="-"))
        date_style.addElement(Day(style="long"))
        doc.styles.addElement(date_style)

        #Integer
        ns1 = NumberStyle(name="IntegerBlack", volatile="true")
        ns1.addElement(Number(decimalplaces="0", minintegerdigits="1", grouping="true"))
        doc.styles.addElement(ns1)

        ns2 = NumberStyle(name="Integer")
        ns2.addElement(TextProperties(color="#ff0000"))
        ns2.addElement(Text(text="-"))
        ns2.addElement(Number(decimalplaces="0", minintegerdigits="1", grouping="true"))
        ns2.addElement(Map(condition="value()>=0", applystylename="IntegerBlack"))
        doc.styles.addElement(ns2)
            
        #Decimal 2
        ns1 = NumberStyle(name="Decimal2Black", volatile="true")
        ns1.addElement(Number(decimalplaces="2", minintegerdigits="1", grouping="true"))
        doc.styles.addElement(ns1)

        ns2 = NumberStyle(name="Decimal2")
        ns2.addElement(TextProperties(color="#ff0000"))
        ns2.addElement(Text(text="-"))
        ns2.addElement(Number(decimalplaces="2", minintegerdigits="1", grouping="true"))
        ns2.addElement(Map(condition="value()>=0", applystylename="Decimal2Black"))
        doc.styles.addElement(ns2)
            
        #Decimal 2
        ns1 = NumberStyle(name="Decimal6Black", volatile="true")
        ns1.addElement(Number(decimalplaces="6", minintegerdigits="1", grouping="true"))
        doc.styles.addElement(ns1)

        ns2 = NumberStyle(name="Decimal6")
        ns2.addElement(TextProperties(color="#ff0000"))
        ns2.addElement(Text(text="-"))
        ns2.addElement(Number(decimalplaces="6", minintegerdigits="1", grouping="true"))
        ns2.addElement(Map(condition="value()>=0", applystylename="Decimal6Black"))
        doc.styles.addElement(ns2)
        
        #Boolean
        ns1 = BooleanStyle(name="BooleanBlack", volatile="true")
        ns1.addElement(Boolean())
        doc.styles.addElement(ns1)

        ns2 = BooleanStyle(name="Boolean")
        ns2.addElement(TextProperties(color="#ff0000"))
        ns2.addElement(Boolean())
        ns2.addElement(Map(condition="value()==1", applystylename="BooleanBlack"))
        doc.styles.addElement(ns2)
            

    def generate_ods_styles(self, doc, currencymanager):
        self.__generate_ods_common_styles(doc)
        for o in self.arr:
            o.generate_ods_styles(doc, currencymanager)


## Class to write Libreoffice Calc ODS files
class ODS_Write(ODS):
    def __init__(self, filename):
        ODS.__init__(self, filename)

        self.colors=ODSStyleColorManager()
        self.colors.append(ODSStyleColor("Green", "#9bff9b", True))
        self.colors.append(ODSStyleColor("GrayDark", "#888888", True))
        self.colors.append(ODSStyleColor("GrayLight", "#bbbbbb", True))
        self.colors.append(ODSStyleColor("Orange", "#ffcc99", True))
        self.colors.append(ODSStyleColor("Yellow", "#ffff7f", True))
        self.colors.append(ODSStyleColor("White", "#ffffff", True))
        self.colors.append(ODSStyleColor("Blue", "#9b9bff", True))
        self.colors.append(ODSStyleColor("Red", "#ff9b9b", True))
        self.colors.append(ODSStyleColor("Normal", "#ffffff", False))
        
        self.currencies=ODSStyleCurrencyManager()
        self.currencies.append(ODSStyleCurrency("EUR", "€"))
        self.currencies.append(ODSStyleCurrency("USD", "$"))

    ## Generate color styles and currency styles and the save file
    def save(self, filename=None):
        self.currencies.generate_ods_styles(self.doc)
        self.colors.generate_ods_styles(self.doc, self.currencies)
        ODS.save(self, filename)
        

## Guess style from color and object class
## @param color_or_style String with a Color name or a style. If it's a color returns the style corresponding to the object. If it's a style returns the same style
## @return String with the style 
def guess_ods_style(color_or_style, object):
    if color_or_style in ["Green", "GrayDark", "GrayLight", "Orange", "Yellow", "White", "Blue", "Red", "Normal"]:
        if object.__class__.__name__=="str":
            return color_or_style + "Left"
        elif object.__class__.__name__=="int":
            return color_or_style + "Integer"
        elif object.__class__.__name__ in ["Currency", "Money" ]:
            return color_or_style + object.currency
        elif object.__class__.__name__=="Percentage":
            return color_or_style + "Percentage"
        elif object.__class__.__name__ in ("Decimal", "float"):
            return color_or_style +  "Decimal2"
        elif object.__class__.__name__=="datetime":
            return color_or_style + "Datetime"
        elif object.__class__.__name__=="date":
            return color_or_style + "Date"
        elif object.__class__.__name__=="time":
            return color_or_style + "Time"
        elif object.__class__.__name__=="bool":
            return color_or_style + "Boolean"
        else:
            info("guess_ods_style not guessed {}".format( object.__class__))
            return "NormalLeft"
    else:
        return color_or_style

## Gets a ODS file and rewrites it with libreoffice convert-to command
## Can be used to assign data values formulas to file. Or to fix ploblems on specific files.
## @param filename_from canbe an ods or a xlsx file.
## @param filename_to Must be a ods file path
## Returns the name of the recently created file
def create_rewritten_ods(filename_from, filename_to=None):  
    if filename_to is None:
        filename_to=filename_from + ".rewritten.ods"
    convert_command(filename_from, filename_to)
    return filename_to

## Creates a new file with data_only cells (formulas converted to numbers).
## Returns the name of the recently created file
## 1 Transforms to xlsx (temporal_xlsx_rewriten)
## 2 Transforms to data_only xlsx (temporal_xlsx_data_only)
def create_data_only_ods( filename_from, filename_to=None):
    if filename_to is None:
        filename_to=filename_from + ".data_only.ods"
    with TemporaryDirectory(prefix="officegenerator_") as tmp_name:
        from officegenerator.libxlsxgenerator import create_data_only_xlsx
        temporal_xlsx=tmp_name+ sep + path.basename(filename_from+".xlsx")
        create_data_only_xlsx(filename_from, temporal_xlsx)
        convert_command(temporal_xlsx, filename_to)
    return filename_to
        
